import sys
from abc import ABC, abstractmethod
from collections.abc import Iterable, Set
from typing import Any, ClassVar, Literal, TypeAlias, TypeVar

from pydantic import model_validator

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    Identifier,
    RequestResource,
    ResponseResource,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

InstanceType: TypeAlias = Literal["node", "edge"]


class TypedInstanceIdentifier(Identifier):
    """Identifier for an Instance instance."""

    instance_type: InstanceType
    space: str
    external_id: str

    def __str__(self) -> str:
        return f"Instance({self.instance_type}, {self.space}, {self.external_id})"

    def dump(self, camel_case: bool = True, include_type: bool = True) -> dict[str, Any]:
        """Dump the resource to a dictionary.

        This is the default serialization method for request resources.
        """
        return self.model_dump(mode="json", by_alias=camel_case, exclude_unset=not include_type)


class TypedNodeIdentifier(TypedInstanceIdentifier):
    instance_type: Literal["node"] = "node"

    def __str__(self) -> str:
        return f"Node({self.space}, {self.external_id})"

    @classmethod
    def from_external_id(cls, item: ExternalId, space: str) -> Self:
        return cls(instance_type="node", space=space, external_id=item.external_id)

    @classmethod
    def from_external_ids(cls, items: Iterable[ExternalId], space: str) -> list[Self]:
        return [cls.from_external_id(item, space) for item in items]

    @classmethod
    def from_str_ids(cls, items: Iterable[str], space: str) -> list[Self]:
        return [cls(instance_type="node", space=space, external_id=item) for item in items]


class TypedEdgeIdentifier(TypedInstanceIdentifier):
    instance_type: Literal["edge"] = "edge"

    def __str__(self) -> str:
        return f"Edge({self.space}, {self.external_id})"


T_TypedInstanceIdentifier = TypeVar("T_TypedInstanceIdentifier", bound=TypedInstanceIdentifier)


class InstanceIdentifier(Identifier):
    space: str
    external_id: str

    def __str__(self) -> str:
        return f"Instance({self.space}, {self.external_id})"


class InstanceResult(BaseModelObject):
    instance_type: InstanceType
    version: int
    was_modified: bool
    space: str
    external_id: str
    created_time: int
    last_updated_time: int

    def as_id(self) -> TypedInstanceIdentifier:
        return TypedInstanceIdentifier(
            instance_type=self.instance_type,
            space=self.space,
            external_id=self.external_id,
        )


class TypedViewReference(Identifier):
    type: Literal["view"] = "view"
    space: str
    external_id: str
    version: str

    def __str__(self) -> str:
        return f"View({self.space}, {self.external_id}, v{self.version})"

    def dump(self, camel_case: bool = True, include_type: bool = True) -> dict[str, Any]:
        """Dump the resource to a dictionary.

        This is the default serialization method for request resources.
        """
        return self.model_dump(mode="json", by_alias=camel_case, exclude_unset=not include_type)

    def as_property_reference(self, property_name: str) -> list[str]:
        return [self.space, f"{self.external_id}/{self.version}", property_name]


######################################################
# The classes below are helper classes for making instances request/responses.
# By using these, we can avoid having to include the instances specific classes in the DTO classes
# that are instance. Instead, these classes can now only have the properties they need to define.
#######################################################


class WrappedInstanceRequest(RequestResource, ABC):
    """This is a base class for resources that are Instances.
    It is used to define resources that are
    """

    VIEW_ID: ClassVar[TypedViewReference]
    instance_type: InstanceType
    space: str
    external_id: str
    existing_version: int | None = None

    def dump(
        self,
        camel_case: bool = True,
        exclude_extra: bool = False,
        context: Literal["api", "toolkit"] = "api",
        exclude: Set[str] | None = None,
    ) -> dict[str, Any]:
        """Dump the resource to a dictionary.

        Args:
            camel_case (bool): Whether to use camelCase for the keys. Default is True.
            exclude_extra (bool): Whether to exclude extra fields not defined in the model. Default is False.
            context (Literal["api", "toolkit"]): The context in which the dump is used. Default is "api".

        """
        exclude_set = set(exclude or set())
        if exclude_extra:
            exclude_set |= set(self.__pydantic_extra__) if self.__pydantic_extra__ else set()
        if context == "api":
            exclude_set.update({"existing_version", "instance_type", "space", "external_id"})
        dumped = self.model_dump(mode="json", by_alias=camel_case, exclude_unset=True, exclude=exclude_set)
        if context == "toolkit":
            return dumped
        return {
            "instanceType": self.instance_type,
            "space": self.space,
            "externalId": self.external_id,
            "sources": [
                {
                    "source": self.VIEW_ID.dump(camel_case=camel_case, include_type=True),
                    "properties": dumped,
                }
            ],
        }


T_WrappedInstanceRequest = TypeVar("T_WrappedInstanceRequest", bound=WrappedInstanceRequest)


class WrappedInstanceResponse(ResponseResource[T_WrappedInstanceRequest], ABC):
    VIEW_ID: ClassVar[TypedViewReference]
    instance_type: InstanceType
    space: str
    external_id: str

    version: int
    created_time: int
    last_updated_time: int
    deleted_time: int | None = None

    @model_validator(mode="before")
    def move_properties(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Move properties from sources to the top level."""
        return move_properties(values, cls.VIEW_ID)

    def dump(
        self, camel_case: bool = True, exclude_extra: bool = False, context: Literal["api", "toolkit"] = "api"
    ) -> dict[str, Any]:
        """Dump the resource to a dictionary.

        Args:
            camel_case (bool): Whether to use camelCase for the keys. Default is True.
            exclude_extra (bool): Whether to exclude extra fields not defined in the model. Default is False.

        """
        output = super().dump(camel_case=camel_case, exclude_extra=exclude_extra)
        if context == "toolkit":
            return output
        properties: dict[str, Any] = {}
        for key in list(output.keys()):
            if key in {
                "instanceType",
                "space",
                "externalId",
                "version",
                "createdTime",
                "lastUpdatedTime",
                "deletedTime",
                "instance_type",
                "external_id",
                "created_time",
                "last_updated_time",
                "deleted_time",
            }:
                continue
            properties[key] = output.pop(key)
        output["properties"] = {
            self.VIEW_ID.space: {
                f"{self.VIEW_ID.external_id}/{self.VIEW_ID.version}": properties,
            }
        }
        return output


def move_properties(values: dict[str, Any], view_id: TypedViewReference) -> dict[str, Any]:
    """Help function to move properties from properties.space.externalId/version to the top level.

    It is used in WrappedInstanceResponse to move properties from the response to the top level.
    """
    if "properties" not in values:
        return values
    properties = values["properties"]
    if not isinstance(properties, dict) or view_id.space not in properties:
        return values
    view_properties = properties[view_id.space]
    identifier = f"{view_id.external_id}/{view_id.version}"
    if not isinstance(view_properties, dict) or identifier not in view_properties:
        return values
    source_properties = view_properties[identifier]
    return {**{key: value for key, value in values.items() if key != "properties"}, **source_properties}


T_WrappedInstanceResponse = TypeVar("T_WrappedInstanceResponse", bound=WrappedInstanceResponse)


class WrappedInstanceListRequest(RequestResource, ABC):
    VIEW_ID: ClassVar[TypedViewReference]
    instance_type: Literal["node"] = "node"
    space: str
    external_id: str

    @abstractmethod
    def dump_instances(self) -> list[dict[str, Any]]:
        """Dumps the object to a list of instance request dictionaries."""
        raise NotImplementedError()

    def as_id(self) -> TypedNodeIdentifier:
        return TypedNodeIdentifier(
            instance_type=self.instance_type,
            space=self.space,
            external_id=self.external_id,
        )

    @abstractmethod
    def as_ids(self) -> list[TypedInstanceIdentifier]:
        """Convert the response to a list of typed instance identifiers."""
        raise NotImplementedError()


T_InstancesListRequest = TypeVar("T_InstancesListRequest", bound=WrappedInstanceListRequest)


class WrappedInstanceListResponse(ResponseResource[T_InstancesListRequest], ABC):
    VIEW_ID: ClassVar[TypedViewReference]
    instance_type: Literal["node"] = "node"
    space: str
    external_id: str

    @model_validator(mode="before")
    @classmethod
    def move_properties(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Move properties from sources to the top level."""
        return move_properties(values, cls.VIEW_ID)

    @abstractmethod
    def as_ids(self) -> list[TypedInstanceIdentifier]:
        """Convert the response to a list of typed instance identifiers."""
        raise NotImplementedError()


T_InstancesListResponse = TypeVar("T_InstancesListResponse", bound=WrappedInstanceListResponse)


class NodeReference(BaseModelObject):
    space: str
    external_id: str



from typing import Literal, TypeAlias

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, RequestResource, ResponseResource
from cognite_toolkit._cdf_tk.client._types import Metadata

from .identifiers import ExternalId

FunctionStatus: TypeAlias = Literal["Queued", "Deploying", "Ready", "Failed", "Retired"]


class FunctionBase(BaseModelObject):
    """Base class for Function with common fields."""

    external_id: str | None = None
    name: str
    file_id: int
    description: str | None = None
    owner: str | None = None
    function_path: str | None = None
    secrets: dict[str, str] | None = None
    env_vars: dict[str, str] | None = None
    cpu: float | None = None
    memory: float | None = None
    runtime: Literal["py38", "py39", "py310", "py311", "py312"] | None = None
    metadata: Metadata | None = None
    index_url: str | None = None
    extra_index_urls: list[str] | None = None


class FunctionRequest(FunctionBase, RequestResource):
    """Request resource for creating/updating functions."""

    def as_id(self) -> ExternalId:
        if self.external_id is None:
            raise ValueError("Cannot create ExternalId: external_id is None")
        return ExternalId(external_id=self.external_id)


class FunctionAPIError(BaseModelObject):
    code: int
    message: str


class FunctionResponse(FunctionBase, ResponseResource[FunctionRequest]):
    """Response resource for functions."""

    id: int
    created_time: int
    runtime_version: str | None = None
    status: FunctionStatus | None = None
    error: FunctionAPIError | None = None

    @classmethod
    def request_cls(cls) -> type[FunctionRequest]:
        return FunctionRequest



from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any
from unittest.mock import MagicMock

from cognite.client.testing import CogniteClientMock
from rich.console import Console

from cognite_toolkit._cdf_tk.client._toolkit_client import ToolkitClient
from cognite_toolkit._cdf_tk.client.api.charts import ChartsAPI
from cognite_toolkit._cdf_tk.client.api.cognite_files import CogniteFilesAPI
from cognite_toolkit._cdf_tk.client.api.containers import ContainersAPI
from cognite_toolkit._cdf_tk.client.api.data_models import DataModelsAPI
from cognite_toolkit._cdf_tk.client.api.graphql_data_models import GraphQLDataModelsAPI
from cognite_toolkit._cdf_tk.client.api.hosted_extractors import HostedExtractorsAPI
from cognite_toolkit._cdf_tk.client.api.legacy.canvas import CanvasAPI, IndustrialCanvasAPI
from cognite_toolkit._cdf_tk.client.api.raw import RawAPI, RawDatabasesAPI, RawTablesAPI
from cognite_toolkit._cdf_tk.client.api.robotics import RoboticsAPI
from cognite_toolkit._cdf_tk.client.api.robotics_capabilities import CapabilitiesAPI
from cognite_toolkit._cdf_tk.client.api.robotics_data_postprocessing import DataPostProcessingAPI
from cognite_toolkit._cdf_tk.client.api.robotics_frames import FramesAPI
from cognite_toolkit._cdf_tk.client.api.robotics_locations import LocationsAPI
from cognite_toolkit._cdf_tk.client.api.robotics_maps import MapsAPI
from cognite_toolkit._cdf_tk.client.api.robotics_robots import RobotsAPI
from cognite_toolkit._cdf_tk.client.api.search_config import SearchConfigurationsAPI
from cognite_toolkit._cdf_tk.client.api.spaces import SpacesAPI
from cognite_toolkit._cdf_tk.client.api.transformation_notifications import TransformationNotificationsAPI
from cognite_toolkit._cdf_tk.client.api.transformation_schedules import TransformationSchedulesAPI
from cognite_toolkit._cdf_tk.client.api.views import ViewsAPI

from ._toolkit_client import ToolAPI
from .api.agents import AgentsAPI
from .api.assets import AssetsAPI
from .api.data_product_versions import DataProductVersionsAPI
from .api.data_products import DataProductsAPI
from .api.datapoint_subscription import DatapointSubscriptionsAPI
from .api.datasets import DataSetsAPI
from .api.events import EventsAPI
from .api.extraction_pipeline_config import ExtractionPipelineConfigsAPI
from .api.extraction_pipelines import ExtractionPipelinesAPI
from .api.filemetadata import FileMetadataAPI
from .api.function_schedules import FunctionSchedulesAPI
from .api.functions import FunctionsAPI
from .api.groups import GroupsAPI
from .api.hosted_extractor_destinations import HostedExtractorDestinationsAPI
from .api.hosted_extractor_jobs import HostedExtractorJobsAPI
from .api.hosted_extractor_mappings import HostedExtractorMappingsAPI
from .api.hosted_extractor_sources import HostedExtractorSourcesAPI
from .api.infield import APMConfigAPI, InfieldAPI, InFieldCDMConfigAPI, InfieldConfigAPI
from .api.instances import InstancesAPI
from .api.labels import LabelsAPI
from .api.location_filters import LocationFiltersAPI
from .api.lookup import (
    AssetLookUpAPI,
    DataSetLookUpAPI,
    EventLookUpAPI,
    ExtractionPipelineLookUpAPI,
    FileMetadataLookUpAPI,
    FunctionLookUpAPI,
    LocationFiltersLookUpAPI,
    LookUpGroup,
    SecurityCategoriesLookUpAPI,
    TimeSeriesLookUpAPI,
)
from .api.migration import (
    CreatedSourceSystemAPI,
    InstanceSourceAPI,
    LookupAPI,
    MigrationAPI,
    MigrationLookupAPI,
    ResourceViewMappingsAPI,
)
from .api.project import ProjectAPI
from .api.relationships import RelationshipsAPI
from .api.security_categories import SecurityCategoriesAPI
from .api.sequence_rows import SequenceRowsAPI
from .api.sequences import SequencesAPI
from .api.simulator_model_revisions import SimulatorModelRevisionsAPI
from .api.simulator_models import SimulatorModelsAPI
from .api.simulator_routine_revisions import SimulatorRoutineRevisionsAPI
from .api.simulator_routines import SimulatorRoutinesAPI
from .api.simulators import SimulatorsAPI
from .api.streamlit_ import StreamlitAPI
from .api.streams import StreamsAPI
from .api.three_d import (
    ThreeDAPI,
    ThreeDClassicAssetMappingAPI,
    ThreeDClassicModelsAPI,
    ThreeDClassicRevisionsAPI,
    ThreeDDMAssetMappingAPI,
)
from .api.timeseries import TimeSeriesAPI
from .api.token import TokenAPI
from .api.transformations import TransformationsAPI
from .api.verify import VerifyAPI
from .api.workflow_triggers import WorkflowTriggersAPI
from .api.workflow_versions import WorkflowVersionsAPI
from .api.workflows import WorkflowsAPI


class ToolkitClientMock(CogniteClientMock):
    """Mock for ToolkitClient object

    All APIs are replaced with specked MagicMock objects.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if "parent" in kwargs:
            super().__init__(*args, **kwargs)
            return None
        super().__init__(*args, **kwargs)
        self.console = Console()
        # Developer note:
        # - Please add your mocked APIs in chronological order
        # - For nested APIs:
        #   - Add spacing above and below
        #   - Use `spec=MyAPI` only for "top level"
        #   - Use `spec_set=MyNestedAPI` for all nested APIs
        self.canvas = MagicMock(spec=CanvasAPI)
        self.canvas.industrial = MagicMock(spec_set=IndustrialCanvasAPI)
        self.charts = MagicMock(spec_set=ChartsAPI)
        self.infield = MagicMock(spec=InfieldAPI)
        self.infield.apm_config = MagicMock(spec_set=APMConfigAPI)
        self.infield.config = MagicMock(spec_set=InfieldConfigAPI)
        self.infield.cdm_config = MagicMock(spec_set=InFieldCDMConfigAPI)

        self.project = MagicMock(spec_set=ProjectAPI)

        self.lookup = MagicMock(spec=LookUpGroup)
        self.lookup.data_sets = MagicMock(spec_set=DataSetLookUpAPI)
        self.lookup.assets = MagicMock(spec_set=AssetLookUpAPI)
        self.lookup.time_series = MagicMock(spec_set=TimeSeriesLookUpAPI)
        self.lookup.files = MagicMock(spec_set=FileMetadataLookUpAPI)
        self.lookup.events = MagicMock(spec_set=EventLookUpAPI)
        self.lookup.security_categories = MagicMock(spec_set=SecurityCategoriesLookUpAPI)
        self.lookup.location_filters = MagicMock(spec_set=LocationFiltersLookUpAPI)
        self.lookup.extraction_pipelines = MagicMock(spec_set=ExtractionPipelineLookUpAPI)
        self.lookup.functions = MagicMock(spec_set=FunctionLookUpAPI)
        self.migration = MagicMock(spec=MigrationAPI)
        self.migration.instance_source = MagicMock(spec_set=InstanceSourceAPI)
        self.migration.lookup = MagicMock(spec=MigrationLookupAPI)
        self.migration.lookup.assets = MagicMock(spec_set=LookupAPI)
        self.migration.lookup.events = MagicMock(spec_set=LookupAPI)
        self.migration.lookup.files = MagicMock(spec_set=LookupAPI)
        self.migration.lookup.time_series = MagicMock(spec_set=LookupAPI)
        self.migration.resource_view_mapping = MagicMock(spec_set=ResourceViewMappingsAPI)
        self.migration.created_source_system = MagicMock(spec_set=CreatedSourceSystemAPI)

        self.tool = MagicMock(spec=ToolAPI)
        self.tool.agents = MagicMock(spec=AgentsAPI)
        self.tool.datapoint_subscriptions = MagicMock(spec=DatapointSubscriptionsAPI)
        self.tool.three_d = MagicMock(spec=ThreeDAPI)
        self.tool.three_d.models_classic = MagicMock(spec_set=ThreeDClassicModelsAPI)
        self.tool.three_d.revisions_classic = MagicMock(spec_set=ThreeDClassicRevisionsAPI)
        self.tool.three_d.asset_mappings_classic = MagicMock(spec_set=ThreeDClassicAssetMappingAPI)
        self.tool.three_d.asset_mappings_dm = MagicMock(spec_set=ThreeDDMAssetMappingAPI)
        self.tool.assets = MagicMock(spec_set=AssetsAPI)
        self.tool.cognite_files = MagicMock(spec_set=CogniteFilesAPI)
        self.tool.timeseries = MagicMock(spec_set=TimeSeriesAPI)
        self.tool.filemetadata = MagicMock(spec_set=FileMetadataAPI)
        self.tool.instances = MagicMock(spec=InstancesAPI)
        self.tool.spaces = MagicMock(spec=SpacesAPI)
        self.tool.views = MagicMock(spec=ViewsAPI)
        self.tool.containers = MagicMock(spec=ContainersAPI)
        self.tool.data_models = MagicMock(spec=DataModelsAPI)
        self.tool.graphql_data_models = MagicMock(spec=GraphQLDataModelsAPI)
        self.tool.location_filters = MagicMock(spec=LocationFiltersAPI)
        self.tool.events = MagicMock(spec_set=EventsAPI)
        self.tool.functions = MagicMock(spec=FunctionsAPI)
        self.tool.functions.schedules = MagicMock(spec_set=FunctionSchedulesAPI)
        self.tool.groups = MagicMock(spec_set=GroupsAPI)
        self.tool.search_configurations = MagicMock(spec_set=SearchConfigurationsAPI)
        self.tool.simulators = MagicMock(spec=SimulatorsAPI)
        self.tool.simulators.models = MagicMock(spec_set=SimulatorModelsAPI)
        self.tool.simulators.model_revisions = MagicMock(spec_set=SimulatorModelRevisionsAPI)
        self.tool.simulators.routines = MagicMock(spec_set=SimulatorRoutinesAPI)
        self.tool.simulators.routine_revisions = MagicMock(spec_set=SimulatorRoutineRevisionsAPI)
        self.tool.datasets = MagicMock(spec_set=DataSetsAPI)
        self.tool.extraction_pipelines = MagicMock(spec=ExtractionPipelinesAPI)
        self.tool.extraction_pipelines.configs = MagicMock(spec_set=ExtractionPipelineConfigsAPI)
        self.tool.hosted_extractors = MagicMock(spec=HostedExtractorsAPI)
        self.tool.hosted_extractors.sources = MagicMock(spec_set=HostedExtractorSourcesAPI)
        self.tool.hosted_extractors.jobs = MagicMock(spec_set=HostedExtractorJobsAPI)
        self.tool.hosted_extractors.destinations = MagicMock(spec_set=HostedExtractorDestinationsAPI)
        self.tool.hosted_extractors.mappings = MagicMock(spec_set=HostedExtractorMappingsAPI)
        self.tool.labels = MagicMock(spec_set=LabelsAPI)
        self.tool.raw = MagicMock(spec=RawAPI)
        self.tool.raw.databases = MagicMock(spec_set=RawDatabasesAPI)
        self.tool.raw.tables = MagicMock(spec_set=RawTablesAPI)
        self.tool.robotics = MagicMock(spec=RoboticsAPI)
        self.tool.robotics.capabilities = MagicMock(spec_set=CapabilitiesAPI)
        self.tool.robotics.data_postprocessing = MagicMock(spec_set=DataPostProcessingAPI)
        self.tool.robotics.frames = MagicMock(spec_set=FramesAPI)
        self.tool.robotics.locations = MagicMock(spec_set=LocationsAPI)
        self.tool.robotics.maps = MagicMock(spec_set=MapsAPI)
        self.tool.robotics.robots = MagicMock(spec_set=RobotsAPI)
        self.tool.relationships = MagicMock(spec_set=RelationshipsAPI)
        self.tool.security_categories = MagicMock(spec_set=SecurityCategoriesAPI)
        self.tool.sequences = MagicMock(spec=SequencesAPI)
        self.tool.sequences.rows = MagicMock(spec_set=SequenceRowsAPI)
        self.tool.transformations = MagicMock(spec=TransformationsAPI)
        self.tool.streamlit = MagicMock(spec=StreamlitAPI)
        self.tool.transformations.schedules = MagicMock(spec=TransformationSchedulesAPI)
        self.tool.transformations.notifications = MagicMock(spec=TransformationNotificationsAPI)
        self.tool.workflows = MagicMock(spec=WorkflowsAPI)
        self.tool.workflows.triggers = MagicMock(spec_set=WorkflowTriggersAPI)
        self.tool.workflows.versions = MagicMock(spec_set=WorkflowVersionsAPI)
        self.tool.data_products = MagicMock(spec=DataProductsAPI)
        self.tool.data_products.versions = MagicMock(spec_set=DataProductVersionsAPI)

        self.streams = MagicMock(spec=StreamsAPI)

        # This is a helper API, not a real API.
        self.token = TokenAPI(self)
        self.verify = MagicMock(spec_set=VerifyAPI)


@contextmanager
def monkeypatch_toolkit_client() -> Iterator[ToolkitClientMock]:
    toolkit_client_mock = ToolkitClientMock()
    try:
        ToolkitClient.__new__ = lambda *args, **kwargs: toolkit_client_mock  # type: ignore[method-assign]
        yield toolkit_client_mock
    finally:
        ToolkitClient.__new__ = lambda cls, *args, **kwargs: object.__new__(cls)  # type: ignore[method-assign]



from pydantic import BaseModel, ConfigDict, DirectoryPath, Field

from cognite_toolkit._cdf_tk.cruds import ResourceTypes
from cognite_toolkit._cdf_tk.resource_classes.base import ToolkitResource

from ._insights import ConsistencyError, ModelSyntaxError, Recommendation
from ._types import AbsoluteFilePath, RelativeDirPath


class BuildVariable(BaseModel):
    id: RelativeDirPath
    value: str | bool | int | float | list[str | bool | int | float]
    is_selected: bool
    iteration: int | None = None

    @property
    def name(self) -> str:
        return self.id.name


class ModuleSource(BaseModel):
    """Class used to describe source for module"""

    path: DirectoryPath = Field(description="Path to the module directory. Can be relative or absolute.")
    id: RelativeDirPath = Field(description="Relative path to the organization directory.")
    resource_files_by_folder: dict[ResourceTypes, list[AbsoluteFilePath]] = Field(default_factory=dict)
    variables: list[BuildVariable] = Field(default_factory=list)
    iteration: int = 0

    @property
    def name(self) -> str:
        return self.path.name


class ResourceType(BaseModel):
    model_config = ConfigDict(frozen=True)

    resource_folder: str
    kind: str


class ReadResource(BaseModel):
    source_path: AbsoluteFilePath


class FailedReadResource(ReadResource):
    errors: list[ModelSyntaxError] = Field(default_factory=list)


class SuccessfulReadResource(ReadResource):
    source_hash: str
    resource_type: ResourceType
    resource: ToolkitResource
    insights: list[Recommendation] = Field(default_factory=list)


class Module(BaseModel):
    """Class used to store module in-memory"""

    source: ModuleSource
    resources: list[ReadResource] = Field(default_factory=list)
    insights: list[Recommendation | ConsistencyError] = Field(default_factory=list)

    @property
    def is_success(self) -> bool:
        return all(isinstance(resource, SuccessfulReadResource) for resource in self.resources)



"""ACL (Access Control Sequence) definitions for Group capabilities.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Groups/operation/createGroups
"""

from collections.abc import Sequence
from typing import Annotated, Any, Literal, TypeAlias

from pydantic import BeforeValidator, Field, JsonValue, TypeAdapter, model_serializer, model_validator
from pydantic_core.core_schema import FieldSerializationInfo

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject
from cognite_toolkit._cdf_tk.client.resource_classes.group._constants import ACL_NAME, SCOPE_NAME
from cognite_toolkit._cdf_tk.utils._auxiliary import get_concrete_subclasses

from .scopes import (
    AllScope,
    AppConfigScope,
    AssetRootIDScope,
    CurrentUserScope,
    DataSetScope,
    ExperimentScope,
    ExtractionPipelineScope,
    IDScope,
    IDScopeLowerCase,
    InstancesScope,
    PartitionScope,
    PostgresGatewayUsersScope,
    Scope,
    SpaceIDScope,
    TableScope,
)


class Acl(BaseModelObject):
    """Base class for all ACL (Access Control Sequence) definitions."""

    acl_name: str
    actions: Sequence[str]
    scope: Scope

    @model_validator(mode="before")
    @classmethod
    def convert_scope_format(cls, value: Any) -> Any:
        """Convert scope from API format {'all': {}} to model format {'scope_name': 'all'}."""
        if not isinstance(value, dict):
            return value
        scope = value.get("scope")
        if not isinstance(scope, dict):
            return value
        # If scope already has scope_name, it's in the correct format
        if SCOPE_NAME in scope:
            return value
        # Convert from API format: find the scope key and extract its data
        scope_name = next(iter(scope.keys()), None)
        if scope_name is not None:
            scope_data = scope.get(scope_name, {})
            if isinstance(scope_data, dict):
                new_scope = {SCOPE_NAME: scope_name, **scope_data}
                value = dict(value)
                value["scope"] = new_scope
        return value

    # MyPy complains that info; FieldSerializationInfo is not compatible with info: Any
    # It is.
    @model_serializer  # type: ignore[type-var]
    def convert_scope_to_api_format(self, info: FieldSerializationInfo) -> dict[str, Any]:
        """Convert scope from model format {'scope_name': 'all'} to API format {'all': {}}."""
        output: dict[str, Any] = {"actions": self.actions}
        scope = self.scope.model_dump(**vars(info))
        if isinstance(scope, dict):
            output["scope"] = {self.scope.scope_name: scope}
        return output


class AgentsAcl(Acl):
    """ACL for Agents resources."""

    acl_name: Literal["agentsAcl"] = Field("agentsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE", "RUN"]]
    scope: AllScope


class AnalyticsAcl(Acl):
    """ACL for Analytics resources."""

    acl_name: Literal["analyticsAcl"] = Field("analyticsAcl", exclude=True)
    actions: Sequence[Literal["READ", "EXECUTE", "LIST"]]
    scope: AllScope


class AnnotationsAcl(Acl):
    """ACL for Annotations resources."""

    acl_name: Literal["annotationsAcl"] = Field("annotationsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE", "SUGGEST", "REVIEW"]]
    scope: AllScope


class AppConfigAcl(Acl):
    """ACL for App Config resources."""

    acl_name: Literal["appConfigAcl"] = Field("appConfigAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | AppConfigScope


class AssetsAcl(Acl):
    """ACL for Assets resources."""

    acl_name: Literal["assetsAcl"] = Field("assetsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class AuditlogAcl(Acl):
    """ACL for Audit Log resources."""

    acl_name: Literal["auditlogAcl"] = Field("auditlogAcl", exclude=True)
    actions: Sequence[Literal["READ"]]
    scope: AllScope


class DataModelInstancesAcl(Acl):
    """ACL for Data Model Instances resources."""

    acl_name: Literal["dataModelInstancesAcl"] = Field("dataModelInstancesAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE", "WRITE_PROPERTIES"]]
    scope: AllScope | SpaceIDScope


class DataModelsAcl(Acl):
    """ACL for Data Models resources."""

    acl_name: Literal["dataModelsAcl"] = Field("dataModelsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | SpaceIDScope


class DataSetsAcl(Acl):
    """ACL for Data Sets resources."""

    acl_name: Literal["datasetsAcl"] = Field("datasetsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE", "OWNER"]]
    scope: AllScope | IDScope


class DiagramParsingAcl(Acl):
    """ACL for Diagram Parsing resources."""

    acl_name: Literal["diagramParsingAcl"] = Field("diagramParsingAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class DigitalTwinAcl(Acl):
    """ACL for Digital Twin resources."""

    acl_name: Literal["digitalTwinAcl"] = Field("digitalTwinAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class DocumentFeedbackAcl(Acl):
    """ACL for Document Feedback resources."""

    acl_name: Literal["documentFeedbackAcl"] = Field("documentFeedbackAcl", exclude=True)
    actions: Sequence[Literal["CREATE", "READ", "DELETE"]]
    scope: AllScope


class DocumentPipelinesAcl(Acl):
    """ACL for Document Pipelines resources."""

    acl_name: Literal["documentPipelinesAcl"] = Field("documentPipelinesAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class EntityMatchingAcl(Acl):
    """ACL for Entity Matching resources."""

    acl_name: Literal["entitymatchingAcl"] = Field("entitymatchingAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class EventsAcl(Acl):
    """ACL for Events resources."""

    acl_name: Literal["eventsAcl"] = Field("eventsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class ExperimentsAcl(Acl):
    """ACL for Experiments resources."""

    acl_name: Literal["experimentAcl"] = Field("experimentAcl", exclude=True)
    actions: Sequence[Literal["USE"]]
    scope: ExperimentScope


class ExtractionConfigsAcl(Acl):
    """ACL for Extraction Configs resources."""

    acl_name: Literal["extractionConfigsAcl"] = Field("extractionConfigsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope | ExtractionPipelineScope


class ExtractionPipelinesAcl(Acl):
    """ACL for Extraction Pipelines resources."""

    acl_name: Literal["extractionPipelinesAcl"] = Field("extractionPipelinesAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | IDScope | DataSetScope


class ExtractionRunsAcl(Acl):
    """ACL for Extraction Runs resources."""

    acl_name: Literal["extractionRunsAcl"] = Field("extractionRunsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope | ExtractionPipelineScope


class FilePipelinesAcl(Acl):
    """ACL for File Pipelines resources."""

    acl_name: Literal["filePipelinesAcl"] = Field("filePipelinesAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class FilesAcl(Acl):
    """ACL for Files resources."""

    acl_name: Literal["filesAcl"] = Field("filesAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class FunctionsAcl(Acl):
    """ACL for Functions resources."""

    acl_name: Literal["functionsAcl"] = Field("functionsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class GeospatialAcl(Acl):
    """ACL for Geospatial resources."""

    acl_name: Literal["geospatialAcl"] = Field("geospatialAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class GeospatialCrsAcl(Acl):
    """ACL for Geospatial CRS resources."""

    acl_name: Literal["geospatialCrsAcl"] = Field("geospatialCrsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class GroupsAcl(Acl):
    """ACL for Groups resources."""

    acl_name: Literal["groupsAcl"] = Field("groupsAcl", exclude=True)
    actions: Sequence[Literal["CREATE", "DELETE", "READ", "LIST", "UPDATE"]]
    scope: AllScope | CurrentUserScope


class HostedExtractorsAcl(Acl):
    """ACL for Hosted Extractors resources."""

    acl_name: Literal["hostedExtractorsAcl"] = Field("hostedExtractorsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class LabelsAcl(Acl):
    """ACL for Labels resources."""

    acl_name: Literal["labelsAcl"] = Field("labelsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class LegacyGenericsAcl(Acl):
    """ACL for Legacy Generics resources."""

    acl_name: Literal["genericsAcl"] = Field("genericsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class LegacyModelHostingAcl(Acl):
    """ACL for Legacy Model Hosting resources."""

    acl_name: Literal["modelHostingAcl"] = Field("modelHostingAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class LocationFiltersAcl(Acl):
    """ACL for Location Filters resources."""

    acl_name: Literal["locationFiltersAcl"] = Field("locationFiltersAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | IDScope


class MonitoringTasksAcl(Acl):
    """ACL for Monitoring Tasks resources."""

    acl_name: Literal["monitoringTasksAcl"] = Field("monitoringTasksAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class NotificationsAcl(Acl):
    """ACL for Notifications resources."""

    acl_name: Literal["notificationsAcl"] = Field("notificationsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class PipelinesAcl(Acl):
    """ACL for Pipelines resources."""

    acl_name: Literal["pipelinesAcl"] = Field("pipelinesAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class PostgresGatewayAcl(Acl):
    """ACL for PostgreSQL Gateway resources."""

    acl_name: Literal["postgresGatewayAcl"] = Field("postgresGatewayAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | PostgresGatewayUsersScope


class ProjectsAcl(Acl):
    """ACL for Projects resources."""

    acl_name: Literal["projectsAcl"] = Field("projectsAcl", exclude=True)
    actions: Sequence[Literal["READ", "CREATE", "LIST", "UPDATE", "DELETE"]]
    scope: AllScope


class RawAcl(Acl):
    """ACL for RAW resources."""

    acl_name: Literal["rawAcl"] = Field("rawAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE", "LIST"]]
    scope: AllScope | TableScope


class RelationshipsAcl(Acl):
    """ACL for Relationships resources."""

    acl_name: Literal["relationshipsAcl"] = Field("relationshipsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class RoboticsAcl(Acl):
    """ACL for Robotics resources."""

    acl_name: Literal["roboticsAcl"] = Field("roboticsAcl", exclude=True)
    actions: Sequence[Literal["READ", "CREATE", "UPDATE", "DELETE"]]
    scope: AllScope | DataSetScope


class SAPWritebackAcl(Acl):
    """ACL for SAP Writeback resources."""

    acl_name: Literal["sapWritebackAcl"] = Field("sapWritebackAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | InstancesScope


class SAPWritebackRequestsAcl(Acl):
    """ACL for SAP Writeback Requests resources."""

    acl_name: Literal["sapWritebackRequestsAcl"] = Field("sapWritebackRequestsAcl", exclude=True)
    actions: Sequence[Literal["WRITE", "LIST"]]
    scope: AllScope | InstancesScope


class ScheduledCalculationsAcl(Acl):
    """ACL for Scheduled Calculations resources."""

    acl_name: Literal["scheduledCalculationsAcl"] = Field("scheduledCalculationsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class SecurityCategoriesAcl(Acl):
    """ACL for Security Categories resources."""

    acl_name: Literal["securityCategoriesAcl"] = Field("securityCategoriesAcl", exclude=True)
    actions: Sequence[Literal["MEMBEROF", "LIST", "CREATE", "UPDATE", "DELETE"]]
    scope: AllScope | IDScopeLowerCase


class SeismicAcl(Acl):
    """ACL for Seismic resources."""

    acl_name: Literal["seismicAcl"] = Field("seismicAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | PartitionScope


class SequencesAcl(Acl):
    """ACL for Sequences resources."""

    acl_name: Literal["sequencesAcl"] = Field("sequencesAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class SessionsAcl(Acl):
    """ACL for Sessions resources."""

    acl_name: Literal["sessionsAcl"] = Field("sessionsAcl", exclude=True)
    actions: Sequence[Literal["LIST", "CREATE", "DELETE"]]
    scope: AllScope


class StreamRecordsAcl(Acl):
    """ACL for Stream Records resources."""

    acl_name: Literal["streamRecordsAcl"] = Field("streamRecordsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | SpaceIDScope


class StreamsAcl(Acl):
    """ACL for Streams resources."""

    acl_name: Literal["streamsAcl"] = Field("streamsAcl", exclude=True)
    actions: Sequence[Literal["READ", "CREATE", "DELETE"]]
    scope: AllScope


class TemplateGroupsAcl(Acl):
    """ACL for Template Groups resources."""

    acl_name: Literal["templateGroupsAcl"] = Field("templateGroupsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class TemplateInstancesAcl(Acl):
    """ACL for Template Instances resources."""

    acl_name: Literal["templateInstancesAcl"] = Field("templateInstancesAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class ThreeDAcl(Acl):
    """ACL for 3D resources."""

    acl_name: Literal["threedAcl"] = Field("threedAcl", exclude=True)
    actions: Sequence[Literal["READ", "CREATE", "UPDATE", "DELETE"]]
    scope: AllScope | DataSetScope


class TimeSeriesAcl(Acl):
    """ACL for Time Series resources."""

    acl_name: Literal["timeSeriesAcl"] = Field("timeSeriesAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope | IDScopeLowerCase | AssetRootIDScope


class TimeSeriesSubscriptionsAcl(Acl):
    """ACL for Time Series Subscriptions resources."""

    acl_name: Literal["timeSeriesSubscriptionsAcl"] = Field("timeSeriesSubscriptionsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class TransformationsAcl(Acl):
    """ACL for Transformations resources."""

    acl_name: Literal["transformationsAcl"] = Field("transformationsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class TypesAcl(Acl):
    """ACL for Types resources."""

    acl_name: Literal["typesAcl"] = Field("typesAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class UserProfilesAcl(Acl):
    """ACL for User Profiles resources."""

    acl_name: Literal["userProfilesAcl"] = Field("userProfilesAcl", exclude=True)
    actions: Sequence[Literal["READ"]]
    scope: AllScope


class VideoStreamingAcl(Acl):
    """ACL for Video Streaming resources."""

    acl_name: Literal["videoStreamingAcl"] = Field("videoStreamingAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE", "SUBSCRIBE", "PUBLISH"]]
    scope: AllScope | DataSetScope


class VisionModelAcl(Acl):
    """ACL for Vision Model resources."""

    acl_name: Literal["visionModelAcl"] = Field("visionModelAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class WellsAcl(Acl):
    """ACL for Wells resources."""

    acl_name: Literal["wellsAcl"] = Field("wellsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope


class WorkflowOrchestrationAcl(Acl):
    """ACL for Workflow Orchestration resources."""

    acl_name: Literal["workflowOrchestrationAcl"] = Field("workflowOrchestrationAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE"]]
    scope: AllScope | DataSetScope


class SimulatorsAcl(Acl):
    """ACL for Simulators resources."""

    acl_name: Literal["simulatorsAcl"] = Field("simulatorsAcl", exclude=True)
    actions: Sequence[Literal["READ", "WRITE", "DELETE", "RUN", "MANAGE"]]
    scope: AllScope | DataSetScope


class UnknownAcl(BaseModelObject):
    """Fallback for unknown ACL types."""

    acl_name: str = Field("unknownAcl", exclude=True)
    actions: Sequence[str]
    scope: dict[str, JsonValue]


def _get_acl_name(cls: type[Acl]) -> str | None:
    """Get the acl_name default value from a Pydantic model class."""
    field = cls.model_fields.get("acl_name")
    if field is not None and field.default is not None:
        return field.default
    return None


_KNOWN_ACLS = {
    name: acl
    for acl in get_concrete_subclasses(Acl)
    if (name := _get_acl_name(acl)) is not None and name != "unknownAcl"
}


def _handle_unknown_acl(value: Any) -> Any:
    if isinstance(value, Acl | UnknownAcl):
        return value
    if isinstance(value, dict) and isinstance(acl_name := value.get(ACL_NAME), str):
        acl_class = _KNOWN_ACLS.get(acl_name)
        if acl_class:
            return TypeAdapter(acl_class).validate_python(value)
    return UnknownAcl.model_validate(value)


AclType: TypeAlias = Annotated[
    (
        AgentsAcl
        | AnalyticsAcl
        | AnnotationsAcl
        | AppConfigAcl
        | AssetsAcl
        | AuditlogAcl
        | DataModelInstancesAcl
        | DataModelsAcl
        | DataSetsAcl
        | DiagramParsingAcl
        | DigitalTwinAcl
        | DocumentFeedbackAcl
        | DocumentPipelinesAcl
        | EntityMatchingAcl
        | EventsAcl
        | ExperimentsAcl
        | ExtractionConfigsAcl
        | ExtractionPipelinesAcl
        | ExtractionRunsAcl
        | FilePipelinesAcl
        | FilesAcl
        | FunctionsAcl
        | GeospatialAcl
        | GeospatialCrsAcl
        | GroupsAcl
        | HostedExtractorsAcl
        | LabelsAcl
        | LegacyGenericsAcl
        | LegacyModelHostingAcl
        | LocationFiltersAcl
        | MonitoringTasksAcl
        | NotificationsAcl
        | PipelinesAcl
        | PostgresGatewayAcl
        | ProjectsAcl
        | RawAcl
        | RelationshipsAcl
        | RoboticsAcl
        | SAPWritebackAcl
        | SAPWritebackRequestsAcl
        | ScheduledCalculationsAcl
        | SecurityCategoriesAcl
        | SeismicAcl
        | SequencesAcl
        | SessionsAcl
        | StreamRecordsAcl
        | StreamsAcl
        | TemplateGroupsAcl
        | TemplateInstancesAcl
        | ThreeDAcl
        | TimeSeriesAcl
        | TimeSeriesSubscriptionsAcl
        | TransformationsAcl
        | TypesAcl
        | UserProfilesAcl
        | VideoStreamingAcl
        | VisionModelAcl
        | WellsAcl
        | WorkflowOrchestrationAcl
        | SimulatorsAcl
        | UnknownAcl
    ),
    BeforeValidator(_handle_unknown_acl),
]



from collections.abc import Hashable, Iterable, Sequence
from typing import Any, final

from cognite.client.data_classes.capabilities import Capability

from cognite_toolkit._cdf_tk.client.resource_classes.data_product_version import (
    DataProductVersionRequest,
    DataProductVersionResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import DataProductVersionId, ExternalId
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.resource_classes import DataProductVersionYAML

from .data_product import DataProductCRUD
from .datamodel import DataModelCRUD


@final
class DataProductVersionCRUD(ResourceCRUD[DataProductVersionId, DataProductVersionRequest, DataProductVersionResponse]):
    folder_name = "data_products"
    resource_cls = DataProductVersionResponse
    resource_write_cls = DataProductVersionRequest
    kind = "DataProductVersion"
    yaml_cls = DataProductVersionYAML
    dependencies = frozenset({DataProductCRUD, DataModelCRUD})
    parent_resource = frozenset({DataProductCRUD})
    support_drop = True
    support_update = True
    _doc_url = "Data-Products/operation/createDataProductVersions"

    @property
    def display_name(self) -> str:
        return "data product versions"

    @classmethod
    def get_id(cls, item: DataProductVersionRequest | DataProductVersionResponse | dict) -> DataProductVersionId:
        if isinstance(item, dict):
            return DataProductVersionId(
                data_product_external_id=item["dataProductExternalId"],
                version=item["version"],
            )
        return item.as_id()

    @classmethod
    def dump_id(cls, id: DataProductVersionId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[DataProductVersionRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        # TODO: dataproductsAcl is not yet in the SDK — return empty to skip capability verification.
        # Once available, require: READ + UPDATE (all version mutations use dataproductsAcl:UPDATE).
        return []

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "dataProductExternalId" in item:
            yield DataProductCRUD, ExternalId(external_id=item["dataProductExternalId"])

    def dump_resource(
        self, resource: DataProductVersionResponse, local: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        dumped["dataProductExternalId"] = resource.data_product_external_id
        local = local or {}
        # The API returns empty objects (e.g. terms: {}) for unset fields; treat as None.
        for key, value in dumped.items():
            if value == {}:
                dumped[key] = None
        defaults: dict[str, Any | None] = {"status": "draft", "description": None, "terms": None}
        for key, default in defaults.items():
            if dumped.get(key) == default and key not in local:
                dumped.pop(key)
        return dumped

    def create(self, items: Sequence[DataProductVersionRequest]) -> list[DataProductVersionResponse]:
        if not items:
            return []
        return self.client.tool.data_products.versions.create(list(items))

    def retrieve(self, ids: Sequence[DataProductVersionId]) -> list[DataProductVersionResponse]:
        if not ids:
            return []
        return self.client.tool.data_products.versions.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[DataProductVersionRequest]) -> list[DataProductVersionResponse]:
        if not items:
            return []
        return self.client.tool.data_products.versions.update(list(items))

    def delete(self, ids: Sequence[DataProductVersionId]) -> int:
        if not ids:
            return 0
        self.client.tool.data_products.versions.delete(list(ids))
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[DataProductVersionResponse]:
        if parent_ids is not None:
            dp_ext_ids = {
                pid.external_id if isinstance(pid, ExternalId) else pid
                for pid in parent_ids
                if isinstance(pid, (str, ExternalId))
            }
        else:
            dp_ext_ids = {dp.external_id for dp in self.client.tool.data_products.list(limit=None)}

        for dp_ext_id in dp_ext_ids:
            for versions in self.client.tool.data_products.versions.iterate(dp_ext_id, limit=None):
                yield from versions



from collections.abc import Iterable, Sequence
from typing import Any

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsSuccessResponse,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import WorkflowVersionId
from cognite_toolkit._cdf_tk.client.resource_classes.workflow_version import (
    WorkflowVersionRequest,
    WorkflowVersionResponse,
)


class WorkflowVersionsAPI(CDFResourceAPI[WorkflowVersionResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "upsert": Endpoint(method="POST", path="/workflows/versions", item_limit=1),
                "retrieve": Endpoint(
                    method="GET", path="/workflow/{workflowExternalId}/versions/{version}", item_limit=1
                ),
                "delete": Endpoint(method="POST", path="/workflows/versions/delete", item_limit=100),
                "list": Endpoint(method="POST", path="/workflows/versions/list", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[WorkflowVersionResponse]:
        return PagedResponse[WorkflowVersionResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[WorkflowVersionId]:
        return ResponseItems[WorkflowVersionId].model_validate_json(response.body)

    def create(self, items: Sequence[WorkflowVersionRequest]) -> list[WorkflowVersionResponse]:
        """Create or update workflow versions in CDF.

        Args:
            items: List of WorkflowVersionRequest objects to create or update.
        Returns:
            List of created/updated WorkflowVersionResponse objects.
        """
        return self._request_item_response(items, "upsert")

    # This is a duplicate of the create method, included to standardize the API interface.
    def update(self, items: Sequence[WorkflowVersionRequest]) -> list[WorkflowVersionResponse]:
        """Create or update workflow versions in CDF.

        Args:
            items: List of WorkflowVersionRequest objects to create or update.
        Returns:
            List of created/updated WorkflowVersionResponse objects.
        """
        return self.create(items)

    def retrieve(
        self, items: Sequence[WorkflowVersionId], ignore_unknown_ids: bool = False
    ) -> list[WorkflowVersionResponse]:
        """Retrieve workflow versions from CDF.

        Args:
            items: List of WorkflowVersionId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        Returns:
            List of retrieved WorkflowVersionResponse objects.
        """
        result: list[WorkflowVersionResponse] = []
        endpoint = self._method_endpoint_map["retrieve"]
        for item in items:
            response = self._http_client.request_single_retries(
                RequestMessage(
                    endpoint_url=self._make_url(f"/workflows/{item.workflow_external_id}/versions/{item.version}"),
                    method=endpoint.method,
                )
            )
            if isinstance(response, SuccessResponse):
                result.append(WorkflowVersionResponse.model_validate_json(response.body))
            elif ignore_unknown_ids:
                continue
            else:
                _ = response.get_success_or_raise()
        return result

    def delete(self, items: Sequence[WorkflowVersionId]) -> None:
        """Delete workflow versions from CDF.

        Args:
            items: List of WorkflowVersionId objects to delete.
        """
        self._request_no_response(items, "delete")

    def _create_list_body(self, workflow_external_id: str | None = None) -> dict[str, Any] | None:
        body: dict[str, Any] = {}
        if workflow_external_id:
            body["filter"] = {"workflowFilters": [{"externalId": workflow_external_id}]}
        return body or None

    def paginate(
        self,
        workflow_external_id: str | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[WorkflowVersionResponse]:
        """Iterate over all workflow versions in CDF.

        Args:
            workflow_external_id: Filter by workflow external ID.
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of WorkflowVersionResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            body=self._create_list_body(workflow_external_id),
        )

    def iterate(
        self,
        workflow_external_id: str | None = None,
        limit: int = 100,
    ) -> Iterable[list[WorkflowVersionResponse]]:
        """Iterate over all workflow versions in CDF.

        Args:
            workflow_external_id: Filter by workflow external ID.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of WorkflowVersionResponse objects.
        """
        return self._iterate(
            limit=limit,
            body=self._create_list_body(workflow_external_id),
        )

    def list(
        self,
        workflow_external_id: str | None = None,
        limit: int | None = 100,
    ) -> list[WorkflowVersionResponse]:
        """List all workflow versions in CDF.

        Returns:
            List of WorkflowVersionResponse objects.
        """
        return self._list(limit=limit, body=self._create_list_body(workflow_external_id))



"""Data Products API for managing CDF data products."""

from collections.abc import Iterable, Sequence
from typing import Literal

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, RequestMessage, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.data_product import DataProductRequest, DataProductResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId

from .data_product_versions import DataProductVersionsAPI


class DataProductsAPI(CDFResourceAPI[DataProductResponse]):
    """API for managing CDF data products."""

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/dataproducts", item_limit=1),
                "retrieve": Endpoint(method="GET", path="/dataproducts/{externalId}", item_limit=1),
                "update": Endpoint(method="POST", path="/dataproducts/update", item_limit=1),
                "delete": Endpoint(method="POST", path="/dataproducts/delete", item_limit=1),
                "list": Endpoint(method="GET", path="/dataproducts", item_limit=1000),
            },
            api_version="alpha",
        )
        self.versions = DataProductVersionsAPI(http_client)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[DataProductResponse]:
        return PagedResponse[DataProductResponse].model_validate_json(response.body)

    def create(self, items: Sequence[DataProductRequest]) -> list[DataProductResponse]:
        """Create data products in CDF.

        Args:
            items: List of DataProductRequest objects to create.
        Returns:
            List of created DataProductResponse objects.

        """
        return self._request_item_response(items, "create")

    def retrieve(
        self, external_ids: Sequence[ExternalId], ignore_unknown_ids: bool = False
    ) -> list[DataProductResponse]:
        """Retrieve data products by external ID.

        The API only supports single-item GET at /dataproducts/{externalId}.

        Args:
            external_ids: List of ExternalId objects of the data products to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs. If False, an error will be raised
                if any of the provided IDs do not exist.
        Returns:
            List of retrieved DataProductResponse objects.

        """
        results: list[DataProductResponse] = []
        endpoint = self._method_endpoint_map["retrieve"]
        for ext_id in external_ids:
            response = self._http_client.request_single_retries(
                RequestMessage(
                    endpoint_url=self._make_url(endpoint.path.format(externalId=ext_id.external_id)),
                    method=endpoint.method,
                    api_version=self._api_version,
                )
            )
            if isinstance(response, SuccessResponse):
                results.append(DataProductResponse.model_validate_json(response.body))
            elif ignore_unknown_ids:
                continue
            else:
                _ = response.get_success_or_raise()
        return results

    def update(
        self, items: Sequence[DataProductRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[DataProductResponse]:
        return self._update(items, mode=mode)

    def delete(self, external_ids: Sequence[ExternalId]) -> None:
        self._request_no_response(external_ids, "delete")

    def paginate(self, limit: int = 10, cursor: str | None = None) -> PagedResponse[DataProductResponse]:
        return self._paginate(cursor=cursor, limit=limit)

    def iterate(self, limit: int | None = 10) -> Iterable[list[DataProductResponse]]:
        return self._iterate(limit=limit)

    def list(self, limit: int | None = 10) -> list[DataProductResponse]:
        return self._list(limit=limit)



from collections.abc import Sequence
from itertools import groupby
from typing import Any, Literal, TypeVar, cast, overload

from cognite.client._api.data_modeling import InstancesAPI
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes.data_modeling import (
    NodeList,
    filters,
    query,
)
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.api.instances import WrappedInstancesAPI
from cognite_toolkit._cdf_tk.client.cdf_client import PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import NodeReference, ViewReference
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import TypedNodeIdentifier
from cognite_toolkit._cdf_tk.client.resource_classes.legacy.migration import (
    AssetCentricId,
    CreatedSourceSystem,
    InstanceSource,
    SpaceSource,
)
from cognite_toolkit._cdf_tk.client.resource_classes.resource_view_mapping import (
    RESOURCE_VIEW_MAPPING_SPACE,
    ResourceViewMappingRequest,
    ResourceViewMappingResponse,
)
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.useful_types import AssetCentricType


class InstanceSourceAPI:
    def __init__(self, instance_api: InstancesAPI) -> None:
        self._instance_api = instance_api
        self._RETRIEVE_LIMIT = 1000
        self._view_id = InstanceSource.get_source()

    def retrieve(self, ids: Sequence[AssetCentricId]) -> NodeList[InstanceSource]:
        """Retrieve a list of instance sources by their IDs.

        Args:
            ids (Sequence[AssetCentricId]): A sequence of AssetCentricId objects representing the IDs of the instance sources to retrieve.
        """
        results: NodeList[InstanceSource] = NodeList[InstanceSource]([])
        for chunk in chunker_sequence(ids, self._RETRIEVE_LIMIT):
            retrieve_query = query.Query(
                with_={
                    "instanceSource": query.NodeResultSetExpression(
                        filter=filters.And(filters.HasData(views=[self._view_id]), self._create_dms_filter(ids)),
                        limit=len(chunk),
                    ),
                },
                select={"instanceSource": query.Select([query.SourceSelector(self._view_id, ["*"])])},
            )
            chunk_response = self._instance_api.query(retrieve_query)
            results.extend([InstanceSource._load(item.dump()) for item in chunk_response.get("instanceSource", [])])
        return results

    @staticmethod
    def _create_dms_filter(ids: Sequence[AssetCentricId]) -> filters.Filter:
        """Create a filter that matches all the AssetCentricIds in the list."""
        if not ids:
            raise ValueError("Cannot create a filter from an empty AssetCentricIdList.")
        to_or_filters: list[filters.Filter] = []
        instance_source_view = InstanceSource.get_source()
        for resource_type, resource_ids in groupby(
            sorted(ids, key=lambda x: x.resource_type), key=lambda x: x.resource_type
        ):
            is_resource = filters.Equals(instance_source_view.as_property_ref("resourceType"), resource_type)
            is_id = filters.In(
                instance_source_view.as_property_ref("id"), [resource_id.id_ for resource_id in resource_ids]
            )
            to_or_filters.append(filters.And(is_resource, is_id))
        return filters.Or(*to_or_filters)

    def list(self, resource_type: AssetCentricType, limit: int | None = DEFAULT_LIMIT_READ) -> NodeList[InstanceSource]:
        """List instance sources filtered by resource type"""
        is_selected = filters.Equals(self._view_id.as_property_ref("resourceType"), resource_type)

        nodes = self._instance_api.list(instance_type="node", filter=is_selected, limit=limit, sources=self._view_id)
        return NodeList[InstanceSource]([InstanceSource._load(node.dump()) for node in nodes])


class ResourceViewMappingsAPI(WrappedInstancesAPI[TypedNodeIdentifier, ResourceViewMappingResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(http_client, ResourceViewMappingRequest.VIEW_ID)

    def _validate_response(self, response: SuccessResponse) -> ResponseItems[TypedNodeIdentifier]:
        return ResponseItems[TypedNodeIdentifier].model_validate_json(response.body)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[ResourceViewMappingResponse]:
        return PagedResponse[ResourceViewMappingResponse].model_validate_json(response.body)

    def list(self, resource_type: str | None = None, limit: int | None = 100) -> list[ResourceViewMappingResponse]:
        filter_: dict[str, Any] = {
            "equals": {
                "property": ["node", "space"],
                "value": RESOURCE_VIEW_MAPPING_SPACE,
            }
        }
        if resource_type:
            filter_ = {
                "and": [
                    filter_,
                    {
                        "equals": {
                            "property": ResourceViewMappingRequest.VIEW_ID.as_property_reference("resourceType"),
                            "value": resource_type,
                        }
                    },
                ]
            }
        return super()._list_instances(filter=filter_, instance_type="node", limit=limit)


class CreatedSourceSystemAPI:
    def __init__(self, instance_api: InstancesAPI) -> None:
        self._instance_api = instance_api
        self._RETRIEVE_LIMIT = 1000
        self._view_id = CreatedSourceSystem.get_source()

    def retrieve(self, source: SequenceNotStr[str]) -> NodeList[CreatedSourceSystem]:
        """Retrieve one or more view sources by their external IDs."""
        results: NodeList[CreatedSourceSystem] = NodeList[CreatedSourceSystem]([])
        # MyPy does not understand that SequenceNotStr is a sequence.
        for chunk in chunker_sequence(source, self._RETRIEVE_LIMIT):  # type: ignore[type-var]
            retrieve_query = query.Query(
                with_={
                    "sourceSystem": query.NodeResultSetExpression(
                        filter=filters.And(filters.HasData(views=[self._view_id]), self._create_dms_filter(chunk)),
                        limit=len(chunk),
                    ),
                },
                select={"sourceSystem": query.Select([query.SourceSelector(self._view_id, ["*"])])},
            )
            chunk_response = self._instance_api.query(retrieve_query)
            results.extend([CreatedSourceSystem._load(item.dump()) for item in chunk_response.get("sourceSystem", [])])
        return results

    def _create_dms_filter(self, source: SequenceNotStr[str]) -> filters.Filter:
        """Create a filter that matches all CreatedSourceSystem with given source in the list."""
        if not source:
            raise ValueError("Cannot create a filter from an empty source list.")
        return filters.In(self._view_id.as_property_ref("source"), list(source))

    def list(self, limit: int = -1) -> NodeList[CreatedSourceSystem]:
        """Lists all created source systems."""
        return self._instance_api.list(instance_type=CreatedSourceSystem, limit=limit)


_T = TypeVar("_T", bound=int | str)


class SpaceSourceAPI:
    def __init__(self, instance_api: InstancesAPI) -> None:
        self._instance_api = instance_api
        self._RETRIEVE_LIMIT = 1000
        self._view_id = SpaceSource.get_source()
        # Cache for space sources by data set ID and external ID
        self._cache_by_id: dict[int, SpaceSource] = {}
        self._cache_by_external_id: dict[str, SpaceSource] = {}

    @overload
    def retrieve(self, data_set_id: int) -> SpaceSource | None: ...

    @overload
    def retrieve(self, data_set_id: Sequence[int]) -> NodeList[SpaceSource]: ...

    @overload
    def retrieve(self, *, data_set_external_id: str) -> SpaceSource | None: ...

    @overload
    def retrieve(self, *, data_set_external_id: SequenceNotStr[str]) -> NodeList[SpaceSource]: ...

    def retrieve(
        self,
        data_set_id: int | Sequence[int] | None = None,
        data_set_external_id: str | SequenceNotStr[str] | None = None,
    ) -> SpaceSource | NodeList[SpaceSource] | None:
        """Retrieve a space source by data set ID or external ID.

        This method uses caching to avoid redundant API calls. If a space source
        has been retrieved before, it will be returned from cache.
        """
        if data_set_id is not None and data_set_external_id is None:
            return self._retrieve_with_cache(
                value=data_set_id,
                property_name="dataSetId",
                cache=self._cache_by_id,
                is_single=isinstance(data_set_id, int),
            )
        elif data_set_external_id is not None:
            return self._retrieve_with_cache(
                # MyPy is confused by SequenceNotStr
                value=data_set_external_id,  # type: ignore[arg-type]
                property_name="dataSetExternalId",
                cache=self._cache_by_external_id,
                is_single=isinstance(data_set_external_id, str),
            )
        else:
            raise ValueError("Either data_set_id or data_set_external_id must be provided, but not both.")

    def _retrieve_with_cache(
        self,
        value: _T | Sequence[_T],
        property_name: str,
        cache: dict[_T, SpaceSource],
        is_single: bool,
    ) -> SpaceSource | NodeList[SpaceSource] | None:
        """Retrieve space sources with caching support."""
        # MyPy does not understand that if is_single is True, value is _T, else Sequence[_T].
        values: list[_T] = [value] if is_single else list(value)  # type: ignore[arg-type, list-item]

        # Check cache for all requested values
        cached_results: list[SpaceSource] = []
        missing_values: list[_T] = []
        for val in values:
            if val in cache:
                cached_results.append(cache[val])
            else:
                missing_values.append(val)

        # Fetch missing values from API
        if missing_values:
            fetched = self._retrieve_by_property(property_name=property_name, values=missing_values)
            for item in fetched:
                self._add_to_cache(item)
                cached_results.append(item)

        if is_single:
            return cached_results[0] if cached_results else None
        return NodeList[SpaceSource](cached_results)

    def _add_to_cache(self, space_source: SpaceSource) -> None:
        """Add a space source to both caches."""
        if space_source.data_set_id:
            self._cache_by_id[space_source.data_set_id] = space_source
        if space_source.data_set_external_id is not None:
            self._cache_by_external_id[space_source.data_set_external_id] = space_source

    def _retrieve_by_property(
        self,
        property_name: str,
        values: Sequence[_T],
    ) -> NodeList[SpaceSource]:
        """Retrieve space sources by filtering on a specific property."""
        results: NodeList[SpaceSource] = NodeList[SpaceSource]([])
        for chunk in chunker_sequence(values, self._RETRIEVE_LIMIT):
            retrieve_query = query.Query(
                with_={
                    "spaceSource": query.NodeResultSetExpression(
                        filter=filters.And(
                            filters.HasData(views=[self._view_id]),
                            filters.In(self._view_id.as_property_ref(property_name), chunk),
                        ),
                        limit=len(chunk),
                    ),
                },
                select={"spaceSource": query.Select([query.SourceSelector(self._view_id, ["*"])])},
            )
            chunk_response = self._instance_api.query(retrieve_query)
            items = chunk_response.get("spaceSource", [])
            results.extend([SpaceSource._load(item.dump()) for item in items])

        return results

    def list(self, limit: int = -1) -> NodeList[SpaceSource]:
        """Lists all space sources and populates the cache.

        Note: This method always fetches from the API and does not use the cache,
        but it will populate the cache with all retrieved space sources.
        """
        results = self._instance_api.list(instance_type=SpaceSource, limit=limit)
        # Populate cache with all retrieved space sources
        for space_source in results:
            self._add_to_cache(space_source)
        return results


_T_Cached = TypeVar("_T_Cached", bound=NodeReference | ViewReference)


class LookupAPI:
    def __init__(self, instance_api: InstancesAPI, resource_type: AssetCentricType) -> None:
        self._instance_api = instance_api
        self._resource_type = resource_type
        self._view_id = InstanceSource.get_source()
        self._node_id_by_id: dict[int, NodeReference | None] = {}
        self._node_id_by_external_id: dict[str, NodeReference | None] = {}
        self._consumer_view_id_by_id: dict[int, ViewReference | None] = {}
        self._consumer_view_id_by_external_id: dict[str, ViewReference | None] = {}
        self._RETRIEVE_LIMIT = 1000

    @overload
    def __call__(self, id: int, external_id: None = None) -> NodeReference | None: ...

    @overload
    def __call__(self, id: SequenceNotStr[int], external_id: None = None) -> dict[int, NodeReference]: ...

    @overload
    def __call__(self, *, external_id: str) -> NodeReference | None: ...

    @overload
    def __call__(self, *, external_id: SequenceNotStr[str]) -> dict[str, NodeReference]: ...

    def __call__(
        self, id: int | SequenceNotStr[int] | None = None, external_id: str | SequenceNotStr[str] | None = None
    ) -> dict[int, NodeReference] | dict[str, NodeReference] | NodeReference | None:
        """Lookup NodeId by either internal ID or external ID.

        Args:
            id (int | Sequence[int] | None): The internal ID(s) to lookup.
            external_id (str | SequenceNotStr[str] | None): The external ID(s) to lookup.

        Returns:
            NodeReference | dict[int, NodeReference] | dict[str, NodeReference] | None: The corresponding NodeReference(s) if found, otherwise None.

        """
        if id is not None and external_id is None:
            return self._lookup(
                identifier=id,
                cache=self._node_id_by_id,
                property_name="id",
                return_type=NodeReference,
                input_type=int,
            )
        elif external_id is not None and id is None:
            return self._lookup(
                identifier=external_id,
                cache=self._node_id_by_external_id,
                property_name="classicExternalId",
                return_type=NodeReference,
                input_type=str,
            )
        else:
            raise TypeError("Either id or external_id must be provided, but not both.")

    @overload
    def consumer_view(self, id: int, external_id: None = None) -> ViewReference | None: ...

    @overload
    def consumer_view(self, id: SequenceNotStr[int], external_id: None = None) -> dict[int, ViewReference]: ...

    @overload
    def consumer_view(self, *, external_id: str) -> ViewReference | None: ...

    @overload
    def consumer_view(self, *, external_id: SequenceNotStr[str]) -> dict[str, ViewReference]: ...

    def consumer_view(
        self, id: int | SequenceNotStr[int] | None = None, external_id: str | SequenceNotStr[str] | None = None
    ) -> dict[int, ViewReference] | dict[str, ViewReference] | ViewReference | None:
        """Lookup Consumer ViewReference by either internal ID or external ID.

        Args:
            id (int | Sequence[int] | None): The internal ID(s) to lookup.
            external_id (str | SequenceNotStr[str] | None): The external ID(s) to lookup.
        Returns:
            ViewReference | dict[int, ViewReference] | dict[str, ViewReference] | None
        """
        if id is not None and external_id is None:
            return self._lookup(
                identifier=id,
                cache=self._consumer_view_id_by_id,
                property_name="id",
                return_type=ViewReference,
                input_type=int,
            )
        elif external_id is not None and id is None:
            return self._lookup(
                identifier=external_id,
                cache=self._consumer_view_id_by_external_id,
                property_name="classicExternalId",
                return_type=ViewReference,
                input_type=str,
            )
        else:
            raise TypeError("Either id or external_id must be provided, but not both.")

    def _lookup(
        self,
        identifier: _T | SequenceNotStr[_T],
        cache: dict[_T, _T_Cached | None],
        property_name: Literal["id", "classicExternalId"],
        return_type: type[_T_Cached],
        input_type: type[_T],
    ) -> dict[_T, _T_Cached] | _T_Cached | None:
        """Generic lookup method for both NodeId and ViewReference by id or external_id."""
        is_single = isinstance(identifier, input_type)
        # MyPy does not understand that if is_single is True, identifier is _T, else SequenceNotStr[_T].
        identifiers: list[_T] = [identifier] if is_single else list(identifier)  # type: ignore[arg-type, list-item]

        missing = [id_ for id_ in identifiers if id_ not in cache]
        if missing:
            self._fetch_and_cache(missing, by=property_name)

        if is_single:
            return cache.get(identifier)  # type: ignore[arg-type]

        return {id_: value for id_ in identifiers if isinstance(value := cache.get(id_), return_type)}

    def _fetch_and_cache(self, identifiers: Sequence[int | str], by: Literal["id", "classicExternalId"]) -> None:
        for chunk in chunker_sequence(identifiers, self._RETRIEVE_LIMIT):
            retrieve_query = query.Query(
                with_={
                    "instanceSource": query.NodeResultSetExpression(
                        filter=filters.And(
                            filters.HasData(views=[self._view_id]),
                            filters.Equals(self._view_id.as_property_ref("resourceType"), self._resource_type),
                            filters.In(self._view_id.as_property_ref(by), list(chunk)),
                        ),
                        limit=len(chunk),
                    ),
                },
                select={"instanceSource": query.Select([query.SourceSelector(self._view_id, ["*"])])},
            )
            chunk_response = self._instance_api.query(retrieve_query)
            items = chunk_response.get("instanceSource", [])
            for item in items:
                instance_source = InstanceSource._load(item.dump())
                node_id = NodeReference(space=instance_source.space, external_id=instance_source.external_id)
                self._node_id_by_id[instance_source.id_] = node_id
                self._consumer_view_id_by_id[instance_source.id_] = instance_source.consumer_view()
                if instance_source.classic_external_id:
                    self._node_id_by_external_id[instance_source.classic_external_id] = node_id
                    self._consumer_view_id_by_external_id[instance_source.classic_external_id] = (
                        instance_source.consumer_view()
                    )
            missing = set(chunk) - set(self._node_id_by_id.keys()) - set(self._node_id_by_external_id.keys())
            if by == "id":
                for missing_id in cast(set[int], missing):
                    if missing_id not in self._node_id_by_id:
                        self._node_id_by_id[missing_id] = None
                    if missing_id not in self._consumer_view_id_by_id:
                        self._consumer_view_id_by_id[missing_id] = None
            elif by == "classicExternalId":
                for missing_ext_id in cast(set[str], missing):
                    if missing_ext_id not in self._node_id_by_external_id:
                        self._node_id_by_external_id[missing_ext_id] = None
                    if missing_ext_id not in self._consumer_view_id_by_external_id:
                        self._consumer_view_id_by_external_id[missing_ext_id] = None


class MigrationLookupAPI:
    def __init__(self, instance_api: InstancesAPI) -> None:
        self.assets = LookupAPI(instance_api, "asset")
        self.events = LookupAPI(instance_api, "event")
        self.files = LookupAPI(instance_api, "file")
        self.time_series = LookupAPI(instance_api, "timeseries")


class MigrationAPI:
    def __init__(self, instance_api: InstancesAPI, http_client: HTTPClient) -> None:
        self.instance_source = InstanceSourceAPI(instance_api)
        self.resource_view_mapping = ResourceViewMappingsAPI(http_client)
        self.created_source_system = CreatedSourceSystemAPI(instance_api)
        self.space_source = SpaceSourceAPI(instance_api)
        self.lookup = MigrationLookupAPI(instance_api)



from abc import ABC
from typing import Annotated, Any, Literal

from pydantic import Field, JsonValue, TypeAdapter, field_serializer
from pydantic_core.core_schema import FieldSerializationInfo

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject

from ._data_types import DataType
from ._references import ContainerDirectReference, ContainerReference, NodeReference, ViewDirectReference, ViewReference


class ViewPropertyDefinition(BaseModelObject, ABC):
    connection_type: str


class ViewCoreProperty(ViewPropertyDefinition, ABC):
    # Core properties do not have connection type in the API, but we add it here such that
    # we can use it as a discriminator in unions. The exclude=True ensures that it is not
    # sent to the API.
    connection_type: Literal["primary_property"] = Field(default="primary_property", exclude=True)
    name: str | None = None
    description: str | None = None
    container: ContainerReference
    container_property_identifier: str

    @field_serializer("container", mode="plain")
    @classmethod
    def serialize_container(cls, container: ContainerReference, info: FieldSerializationInfo) -> dict[str, Any]:
        return {**container.model_dump(**vars(info)), "type": "container"}


class ViewCorePropertyRequest(ViewCoreProperty):
    source: ViewReference | None = None

    @field_serializer("source", mode="plain")
    @classmethod
    def serialize_source(cls, source: ViewReference | None, info: FieldSerializationInfo) -> dict[str, Any] | None:
        if source is None:
            return None
        return {**source.model_dump(**vars(info)), "type": "view"}


class ConstraintOrIndexState(BaseModelObject):
    nullability: Literal["current", "pending", "failed"] | None = None
    max_list_size: Literal["current", "pending", "failed"] | None = None
    max_text_size: Literal["current", "pending", "failed"] | None = None


class ViewCorePropertyResponse(ViewCoreProperty):
    immutable: bool | None = None
    nullable: bool | None = None
    auto_increment: bool | None = None
    default_value: str | int | bool | dict[str, JsonValue] | None = None
    constraint_state: ConstraintOrIndexState
    type: DataType

    def as_request(self) -> ViewCorePropertyRequest:
        return ViewCorePropertyRequest.model_validate(self.model_dump(by_alias=True))


class ConnectionPropertyDefinition(ViewPropertyDefinition, ABC):
    name: str | None = None
    description: str | None = None


class EdgeProperty(ConnectionPropertyDefinition, ABC):
    source: ViewReference
    type: NodeReference
    edge_source: ViewReference | None = None
    direction: Literal["outwards", "inwards"] = "outwards"

    @field_serializer("source", "edge_source", mode="plain")
    @classmethod
    def serialize_source(cls, source: ViewReference | None, info: FieldSerializationInfo) -> dict[str, Any] | None:
        if source is None:
            return None
        return {**source.model_dump(**vars(info)), "type": "view"}


class SingleEdgeProperty(EdgeProperty):
    connection_type: Literal["single_edge_connection"] = "single_edge_connection"


class MultiEdgeProperty(EdgeProperty):
    connection_type: Literal["multi_edge_connection"] = "multi_edge_connection"


class ReverseDirectRelationProperty(ConnectionPropertyDefinition, ABC):
    source: ViewReference
    through: ContainerDirectReference | ViewDirectReference

    @field_serializer("source", mode="plain")
    @classmethod
    def serialize_source(cls, source: ViewReference, info: FieldSerializationInfo) -> dict[str, Any]:
        return {**source.model_dump(**vars(info)), "type": "view"}

    @field_serializer("through", mode="plain")
    @classmethod
    def serialize_through(
        cls, through: ContainerDirectReference | ViewDirectReference, info: FieldSerializationInfo
    ) -> dict[str, Any]:
        output = through.model_dump(**vars(info))
        if isinstance(through, ContainerDirectReference):
            output["source"]["type"] = "container"
        else:
            output["source"]["type"] = "view"
        return output


class SingleReverseDirectRelationPropertyRequest(ReverseDirectRelationProperty):
    connection_type: Literal["single_reverse_direct_relation"] = "single_reverse_direct_relation"


class MultiReverseDirectRelationPropertyRequest(ReverseDirectRelationProperty):
    connection_type: Literal["multi_reverse_direct_relation"] = "multi_reverse_direct_relation"


class SingleReverseDirectRelationPropertyResponse(ReverseDirectRelationProperty):
    connection_type: Literal["single_reverse_direct_relation"] = "single_reverse_direct_relation"
    targets_list: bool

    def as_request(self) -> SingleReverseDirectRelationPropertyRequest:
        return SingleReverseDirectRelationPropertyRequest.model_validate(self.model_dump(by_alias=True))


class MultiReverseDirectRelationPropertyResponse(ReverseDirectRelationProperty):
    connection_type: Literal["multi_reverse_direct_relation"] = "multi_reverse_direct_relation"
    targets_list: bool

    def as_request(self) -> MultiReverseDirectRelationPropertyRequest:
        return MultiReverseDirectRelationPropertyRequest.model_validate(self.model_dump(by_alias=True))


ViewRequestProperty = Annotated[
    SingleEdgeProperty
    | MultiEdgeProperty
    | SingleReverseDirectRelationPropertyRequest
    | MultiReverseDirectRelationPropertyRequest
    | ViewCorePropertyRequest,
    Field(discriminator="connection_type"),
]
ViewResponseProperty = Annotated[
    SingleEdgeProperty
    | MultiEdgeProperty
    | SingleReverseDirectRelationPropertyResponse
    | MultiReverseDirectRelationPropertyResponse
    | ViewCorePropertyResponse,
    Field(discriminator="connection_type"),
]

ViewRequestPropertyAdapter: TypeAdapter[ViewRequestProperty] = TypeAdapter(ViewRequestProperty)



from pydantic import Field

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    RequestResource,
    ResponseResource,
)

from .identifiers import InternalIdUnwrapped


class SecurityCategory(BaseModelObject):
    name: str


class SecurityCategoryRequest(SecurityCategory, RequestResource):
    # This is not part of the request payload.
    id: int | None = Field(None, exclude=True)

    def as_id(self) -> InternalIdUnwrapped:
        if self.id is None:
            raise ValueError("Cannot create identifier from SecurityCategoryRequest without id set")
        return InternalIdUnwrapped(id=self.id)


class SecurityCategoryResponse(SecurityCategory, ResponseResource[SecurityCategoryRequest]):
    id: int

    @classmethod
    def request_cls(cls) -> type[SecurityCategoryRequest]:
        return SecurityCategoryRequest

    def as_id(self) -> InternalIdUnwrapped:
        return InternalIdUnwrapped(id=self.id)



from typing import Annotated, Any, Literal, TypeAlias

from pydantic import Field, JsonValue, field_serializer, field_validator
from pydantic_core.core_schema import ValidationInfo

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    RequestResource,
    ResponseResource,
)

from .identifiers import Identifier, WorkflowVersionId

TaskType: TypeAlias = Literal["function", "transformation", "cdf", "dynamic", "subworkflow", "simulation"]


class TaskId(Identifier):
    external_id: str

    def __str__(self) -> str:
        return f"externalId='{self.external_id}'"


class TaskParameterDefinition(BaseModelObject):
    type: str


class CogniteFunctionRef(BaseModelObject):
    external_id: str
    data: JsonValue | str | None = None


class FunctionTaskParameters(TaskParameterDefinition):
    type: Literal["function"] = Field("function", exclude=True)
    function: CogniteFunctionRef
    is_async_complete: bool | None = None


class TransformationRef(BaseModelObject):
    external_id: str
    concurrency_policy: Literal["fail", "waitForCurrent", "restartAfterCurrent"] | None = None
    use_transformation_credentials: bool | None = None


class TransformationTaskParameters(TaskParameterDefinition):
    type: Literal["transformation"] = Field("transformation", exclude=True)
    transformation: TransformationRef


class CDFRequest(BaseModelObject):
    resource_path: str
    method: str
    query_parameters: JsonValue | str | None = None
    body: JsonValue | str | None = None
    request_timeout_in_millis: float | str | None = None
    cdf_version_header: str | None = None


class CDFTaskParameters(TaskParameterDefinition):
    type: Literal["cdf"] = Field("cdf", exclude=True)
    cdf_request: CDFRequest


class DynamicRef(BaseModelObject):
    tasks: str


class DynamicTaskParameters(TaskParameterDefinition):
    type: Literal["dynamic"] = Field("dynamic", exclude=True)
    dynamic: DynamicRef


class SubworkflowRef(BaseModelObject):
    tasks: "list[Task]"


class SubworkflowTaskParameters(TaskParameterDefinition):
    type: Literal["subworkflow"] = Field("subworkflow", exclude=True)
    subworkflow: WorkflowVersionId | SubworkflowRef


class SimulatorInputUnit(BaseModelObject):
    name: str


class SimulatorInput(BaseModelObject):
    reference_id: str
    value: str | int | float | list[str] | list[int] | list[float]
    unit: SimulatorInputUnit | None = None


class SimulationRef(BaseModelObject):
    routine_external_id: str
    run_time: int | None = None
    inputs: list[SimulatorInput] | None = None


class SimulationTaskParameters(TaskParameterDefinition):
    type: Literal["simulation"] = Field("simulation", exclude=True)
    simulation: SimulationRef


Parameter = Annotated[
    FunctionTaskParameters
    | TransformationTaskParameters
    | CDFTaskParameters
    | DynamicTaskParameters
    | SubworkflowTaskParameters
    | SimulationTaskParameters,
    Field(discriminator="type"),
]


class Task(BaseModelObject):
    external_id: str
    type: TaskType
    name: str | None = None
    description: str | None = None
    retries: int | None = None
    timeout: int | None = None
    on_failure: Literal["abortWorkflow", "skipTask"] | None = None
    depends_on: list[TaskId] | None = None
    parameters: Parameter

    @field_validator("parameters", mode="before")
    @classmethod
    def move_type_to_field(cls, value: Any, info: ValidationInfo) -> Any:
        if not isinstance(value, dict) or "type" not in info.data:
            return value
        value = dict(value)
        value["type"] = info.data["type"]
        return value

    @field_serializer("type", mode="plain")
    def prefer_parameter_type(self, value: Any) -> Any:
        # In case the type is set by accident to not match parameters.type
        # we prefer parameters.type
        return self.parameters.type


class WorkflowDefinition(BaseModelObject):
    description: str | None = None
    tasks: list[Task]


class WorkflowVersion(BaseModelObject):
    workflow_external_id: str
    version: str
    workflow_definition: WorkflowDefinition

    def as_id(self) -> WorkflowVersionId:
        return WorkflowVersionId(workflow_external_id=self.workflow_external_id, version=self.version)


class WorkflowVersionRequest(WorkflowVersion, RequestResource): ...


class WorkflowVersionResponse(WorkflowVersion, ResponseResource[WorkflowVersionRequest]):
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[WorkflowVersionRequest]:
        return WorkflowVersionRequest



from typing import Annotated, Any, Literal

from pydantic import Field, JsonValue, field_validator

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)

from .identifiers import ExternalId


class JobFormatDefinition(BaseModelObject):
    type: str


class PrefixConfig(BaseModelObject):
    from_topic: bool | None = None
    prefix: str | None = None


class SpaceRef(BaseModelObject):
    space: str


class CogniteFormat(JobFormatDefinition):
    type: Literal["cognite"] = "cognite"
    encoding: str | None = None
    compression: str | None = None
    prefix: PrefixConfig | None = None
    data_models: list[SpaceRef] | None = None


class CustomFormat(JobFormatDefinition):
    type: Literal["custom"] = "custom"
    encoding: str | None = None
    compression: str | None = None
    mapping_id: str


class RockwellFormat(JobFormatDefinition):
    type: Literal["rockwell"] = "rockwell"
    encoding: str | None = None
    compression: str | None = None
    prefix: PrefixConfig | None = None
    data_models: list[SpaceRef] | None = None


class ValueFormat(JobFormatDefinition):
    type: Literal["value"] = "value"
    encoding: str | None = None
    compression: str | None = None
    prefix: PrefixConfig | None = None
    data_models: list[SpaceRef] | None = None


JobFormat = Annotated[
    CogniteFormat | CustomFormat | RockwellFormat | ValueFormat,
    Field(discriminator="type"),
]


class IncrementalLoadDefinition(BaseModelObject):
    type: str


class BodyIncrementalLoad(IncrementalLoadDefinition):
    type: Literal["body"] = "body"
    value: str


class NextUrlIncrementalLoad(IncrementalLoadDefinition):
    type: Literal["nextUrl"] = "nextUrl"
    value: str


class HeaderIncrementalLoad(IncrementalLoadDefinition):
    type: Literal["headerValue"] = "headerValue"
    key: str
    value: str


class QueryParamIncrementalLoad(IncrementalLoadDefinition):
    type: Literal["queryParam"] = "queryParam"
    key: str
    value: str


class MQTTConfig(BaseModelObject):
    topic_filter: str


class KafkaConfig(BaseModelObject):
    topic: str
    partitions: int | None = None


class RestConfig(BaseModelObject):
    interval: Literal["5m", "15m", "1h", "6h", "12h", "1d"]
    path: str
    method: Literal["get", "post"] | None = None
    body: JsonValue | None = None
    query: dict[str, str] | None = None
    headers: dict[str, str] | None = None
    incremental_load: BodyIncrementalLoad | HeaderIncrementalLoad | QueryParamIncrementalLoad | None = Field(
        None, discriminator="type"
    )
    pagination: (
        BodyIncrementalLoad | NextUrlIncrementalLoad | HeaderIncrementalLoad | QueryParamIncrementalLoad | None
    ) = Field(None, discriminator="type")


class HostedExtractorJob(BaseModelObject):
    external_id: str
    destination_id: str
    source_id: str
    format: JobFormat
    config: MQTTConfig | KafkaConfig | RestConfig | None = None

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)

    @field_validator("config", mode="before")
    @classmethod
    def empty_dict_as_none(cls, value: Any) -> Any:
        if value == {}:
            return None
        return value


class HostedExtractorJobRequest(HostedExtractorJob, UpdatableRequestResource):
    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, Any]:
        update_item = super().as_update(mode=mode)
        exclude_unset = mode == "patch"
        update_item["update"]["format"] = {
            "set": self.format.model_dump(mode="json", exclude_none=True, by_alias=True, exclude_unset=exclude_unset)
        }
        return update_item


class HostedExtractorJobResponse(HostedExtractorJob, ResponseResource[HostedExtractorJobRequest]):
    status: str | None = None
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[HostedExtractorJobRequest]:
        return HostedExtractorJobRequest



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import ClassVar, Generic, TypeVar

from cognite_toolkit._cdf_tk.commands.build_v2.data_classes._insights import ConsistencyError, Recommendation
from cognite_toolkit._cdf_tk.commands.build_v2.data_classes._module import Module
from cognite_toolkit._cdf_tk.resource_classes.base import ToolkitResource

T_Resource = TypeVar("T_Resource", bound=ToolkitResource)
T_Resources = TypeVar("T_Resources", bound=Sequence[ToolkitResource] | Module)


class ToolkitRule(ABC, Generic[T_Resources]):
    """A base rule class that defines the structure for all validation rules in the toolkit.
    It is defined in a such way to enable two different types of subclasses:
    - rules that validate individual resources (e.g. DataModelRule)
    - and rules that validate across resources in a module (e.g. ModuleRule).

    """

    code: ClassVar[str]
    resource_type: ClassVar[str]
    insight_type: ClassVar[type[ConsistencyError] | type[Recommendation]]
    alpha: ClassVar[bool] = False
    fixable: ClassVar[bool] = False

    def __init__(
        self,
        resources: T_Resources,
    ) -> None:
        self.resources: T_Resources = resources

    @abstractmethod
    def validate(self) -> list[ConsistencyError] | list[Recommendation] | list[ConsistencyError | Recommendation]:
        """Execute rule validation."""
        ...


class ToolkitResourceRule(Generic[T_Resource], ToolkitRule[Sequence[T_Resource]]):
    """Rule for toolkit resource validation principles."""

    pass


class ToolkitModuleRule(ToolkitRule[Module]):
    """Rule for toolkit module validation principles."""

    pass



from typing import Literal

from cognite.client.data_classes.data_modeling import ContainerId
from pydantic import Field

from cognite_toolkit._cdf_tk.constants import DM_EXTERNAL_ID_PATTERN, SPACE_FORMAT_PATTERN

from ._base import DataSelector, SelectorObject


class SelectedStream(SelectorObject):
    """Selected stream for records."""

    external_id: str


class SelectedContainer(SelectorObject):
    """Selected container for records."""

    space: str = Field(
        description="Id of the space that the container belongs to.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the container.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )

    def as_id(self) -> ContainerId:
        return ContainerId(space=self.space, external_id=self.external_id)

    def __str__(self) -> str:
        return f"{self.space}_{self.external_id}"


class RecordContainerSelector(DataSelector):
    type: Literal["recordContainer"] = "recordContainer"
    kind: Literal["Records"] = "Records"
    stream: SelectedStream
    container: SelectedContainer

    def __str__(self) -> str:
        return f"{self.container.space}_{self.container.external_id}"



from typing import Annotated, Any, TypeAlias

from pydantic import BeforeValidator


def _keys_and_values_as_string(value: Any) -> Any:
    if not isinstance(value, dict):
        return value
    return {str(k): str(v) for k, v in value.items()}


Metadata: TypeAlias = Annotated[dict[str, str], BeforeValidator(_keys_and_values_as_string)]



import hashlib
from abc import ABC
from typing import Literal

from ._base import DataSelector


class CanvasSelector(DataSelector, ABC):
    kind: Literal["IndustrialCanvas"] = "IndustrialCanvas"


class CanvasExternalIdSelector(CanvasSelector):
    type: Literal["canvasExternalId"] = "canvasExternalId"
    external_ids: tuple[str, ...]

    def __str__(self) -> str:
        hash_ = hashlib.md5(",".join(sorted(self.external_ids)).encode()).hexdigest()[:8]
        return f"canvas_count_{len(self.external_ids)}_hash_{hash_}"



import os
import sys
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Any

from pydantic import JsonValue, TypeAdapter, ValidationError
from rich.console import Console
from rich.panel import Panel

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.commands.build_v2._module_source_parser import ModuleSourceParser
from cognite_toolkit._cdf_tk.commands.build_v2.data_classes import (
    BuildFolder,
    BuildParameters,
    BuildSourceFiles,
    BuiltModule,
    ConfigYAML,
    Module,
    ModuleSource,
    RelativeDirPath,
    ResourceType,
    ValidationType,
)
from cognite_toolkit._cdf_tk.commands.build_v2.data_classes._insights import ModelSyntaxError, Recommendation
from cognite_toolkit._cdf_tk.commands.build_v2.data_classes._module import (
    BuildVariable,
    FailedReadResource,
    ReadResource,
    SuccessfulReadResource,
)
from cognite_toolkit._cdf_tk.commands.build_v2.data_classes._plugins import NeatPlugin
from cognite_toolkit._cdf_tk.commands.build_v2.data_classes._types import AbsoluteFilePath
from cognite_toolkit._cdf_tk.constants import HINT_LEAD_TEXT, MODULES
from cognite_toolkit._cdf_tk.cruds import RESOURCE_CRUD_BY_FOLDER_NAME, ResourceCRUD
from cognite_toolkit._cdf_tk.cruds._resource_cruds.datamodel import DataModelCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitFileNotFoundError, ToolkitNotADirectoryError, ToolkitValueError
from cognite_toolkit._cdf_tk.resource_classes import ToolkitResource
from cognite_toolkit._cdf_tk.rules import RulesOrchestrator
from cognite_toolkit._cdf_tk.utils import calculate_hash, humanize_collection, safe_write, sanitize_filename
from cognite_toolkit._cdf_tk.utils.file import read_yaml_content, relative_to_if_possible, safe_read, yaml_safe_dump
from cognite_toolkit._cdf_tk.validation import humanize_validation_error


class BuildV2Command(ToolkitCommand):
    def build(self, parameters: BuildParameters, client: ToolkitClient | None = None) -> BuildFolder:
        console = client.console if client else Console()

        self._validate_build_parameters(parameters, console, sys.argv)
        build_files = self._read_file_system(parameters)
        module_sources = self._parse_module_sources(build_files)

        build_folder = self._build_modules(module_sources, parameters.build_dir)

        # Todo: Some mixpanel tracking.
        # Can be parallelized with number of plugins.
        # Neat is done inside the global validation.
        self._global_validation(build_folder, client)

        self._write_results(build_folder)
        return build_folder

    @classmethod
    def _validate_build_parameters(cls, parameters: BuildParameters, console: Console, user_args: list[str]) -> None:
        """Checks that the user has the correct folders set up and that the config file (if provided) exists."""

        # Set up the variables
        organization_dir = parameters.organization_dir
        organization_dir_display = relative_to_if_possible(organization_dir)
        config_yaml_path: Path | None = None
        if parameters.config_yaml_name:
            config_yaml_path = organization_dir / ConfigYAML.get_filename(parameters.config_yaml_name)
            content = f"  ┣ {MODULES}/\n  ┗ {config_yaml_path.name}\n"
        else:
            content = f"  ┗ {MODULES}\n"
        expected_panel = Panel(
            f"Toolkit expects the following structure:\n{organization_dir_display.as_posix()!r}/\n{content}",
            expand=False,
        )
        module_directory = parameters.modules_directory

        # Execute the checks.
        if module_directory.exists() and (config_yaml_path is None or config_yaml_path.exists()):
            # All good.
            return

        console.print(expected_panel)
        if not organization_dir.exists():
            raise ToolkitNotADirectoryError(f"Organization directory '{organization_dir_display.as_posix()}' not found")
        elif module_directory.exists() and config_yaml_path is not None and not config_yaml_path.exists():
            raise ToolkitFileNotFoundError(
                f"Config YAML file '{relative_to_if_possible(config_yaml_path).as_posix()}' not found"
            )
        elif not module_directory.exists():
            # This is likely the most common error. The user pass in Path.cwd() as the organization directory,
            # but the actual organization directory is a subdirectory.
            candidate_org = next(
                (subdir for subdir in Path.cwd().iterdir() if subdir.is_dir() and (subdir / MODULES).exists()), None
            )
            if candidate_org:
                display_path = relative_to_if_possible(candidate_org)
                suggested_command = cls._create_suggested_command(display_path, user_args)
                console.print(
                    f"\n{HINT_LEAD_TEXT}Did [red]you[/red] mean to use the command {suggested_command}?\n",
                    markup=True,
                )
                cdf_toml = CDFToml.load()
                if not cdf_toml.cdf.has_user_set_default_org:
                    console.print(
                        f"{HINT_LEAD_TEXT} You can specify a 'default_organization_dir = ...' in the 'cdf' section of your "
                        f"'{CDFToml.file_name}' file to avoid using the -o/--organization-dir argument",
                        markup=True,
                    )
            raise ToolkitNotADirectoryError(
                f"Could not find the modules directory at '{relative_to_if_possible(module_directory).as_posix()}'"
            )
        else:
            raise NotImplementedError("Unhandled case. Please report this.")

    @classmethod
    def _create_suggested_command(cls, display_path: Path, user_args: list[str]) -> str:
        suggestion = ["cdf"]
        skip_next = False
        found = False
        for arg in user_args[1:]:
            if skip_next:
                skip_next = False
                continue

            arg_name = arg.split("=")[0]
            if arg_name in ("-o", "--organization-dir"):
                suggestion.append(f"{arg_name} {display_path}")
                found = True
                if "=" not in arg:
                    skip_next = True
                continue
            suggestion.append(arg)
        if not found:
            suggestion.append(f"-o {display_path}")
        return f"'{' '.join(suggestion)}'"

    def _parse_module_sources(self, build: BuildSourceFiles) -> list[ModuleSource]:
        parser = ModuleSourceParser(build.selected_modules, build.organization_dir)
        module_sources = parser.parse(build.yaml_files, build.variables)
        if parser.errors:
            # Todo: Nicer way of formatting errors. Jira CDF-27107
            raise ToolkitValueError(
                "Errors encountered while parsing modules:\n" + "\n".join(f"- {error!s}" for error in parser.errors)
            )
        return module_sources

    @classmethod
    def _read_file_system(cls, parameters: BuildParameters) -> BuildSourceFiles:
        """Reads the file system to find the YAML files to build along with config.<name>.yaml if it exists."""
        selected: set[RelativeDirPath | str] = {
            parameters.modules_directory.relative_to(parameters.organization_dir)
        }  # Default to everything under modules.
        variables: dict[str, JsonValue] = {}
        cdf_project: str = os.environ.get("CDF_PROJECT", "UNKNOWN")
        validation_type: ValidationType = "prod"
        if parameters.user_selected_modules:
            selected, errors = cls._parse_user_selection(parameters.user_selected_modules, parameters.organization_dir)
            if errors:
                raise ToolkitValueError("Invalid module selection:\n" + "\n".join(f"- {error}" for error in errors))

        if parameters.config_yaml_name:
            try:
                config = ConfigYAML.from_yaml_file(
                    ConfigYAML.get_filepath(parameters.organization_dir, parameters.config_yaml_name)
                )
            except ValidationError as e:
                errors = humanize_validation_error(e)
                raise ToolkitValueError(
                    f"Config YAML file '{parameters.config_yaml_name}' is invalid:\n{'- '.join(errors)}"
                ) from e
            if not parameters.user_selected_modules and config.environment.selected:
                selected, errors = cls._parse_user_selection(config.environment.selected, parameters.organization_dir)
                if errors:
                    raise ToolkitValueError("Invalid module selection:\n" + "\n".join(f"- {error}" for error in errors))
            variables = config.variables or {}
            cdf_project = config.environment.project
            validation_type = config.environment.validation_type

        # Todo optimize by only searching for yaml files in the selected modules paths if selection is provided.
        yaml_files = [
            yaml_file.relative_to(parameters.organization_dir)
            for yaml_file in parameters.modules_directory.rglob("*.y*ml")
        ]
        return BuildSourceFiles(
            yaml_files=yaml_files,
            selected_modules=selected,
            variables=variables,
            validation_type=validation_type,
            cdf_project=cdf_project,
            organization_dir=parameters.organization_dir,
        )

    @classmethod
    def _parse_user_selection(
        cls, user_selected_modules: list[str], organization_dir: Path
    ) -> tuple[set[RelativeDirPath | str], list[str]]:
        selected: set[RelativeDirPath | str] = set()
        errors: list[str] = []
        for item in user_selected_modules:
            if "/" not in item:
                # Module name provided
                selected.add(item)
                continue

            item_path = Path(item)
            if not Path(item).resolve().is_relative_to(organization_dir):
                errors.append(f"Selected module path {item_path.as_posix()!r} is not under the organization directory")
                continue

            if item_path.is_absolute():
                errors.append(
                    f"Selected module path {item_path.as_posix()!r} should be relative to the organization directory"
                )
                continue
            if not (organization_dir / item_path).exists():
                errors.append(
                    f"Selected module path {item_path.as_posix()!r} does not exist under the organization directory"
                )
                continue
            if not item_path.is_dir():
                errors.append(f"Selected module path {item_path.as_posix()!r} is not a directory")
                continue
            selected.add(item_path)
        return selected, errors

    def _build_modules(
        self, module_sources: Sequence[ModuleSource], build_dir: Path, max_workers: int = 1
    ) -> BuildFolder:
        folder: BuildFolder = BuildFolder(path=build_dir)
        for source in module_sources:
            # Inside this loop, do not raise exceptions.
            module = self._import_module(source)  # Syntax validation

            # init built_module
            built_module = BuiltModule(source=module.source)

            if module.is_success:
                self._local_validation(module)
                built_module.built_files = self._export_module(module, build_dir)

            built_module.insights.extend(module.insights)
            for resource in module.resources:
                if isinstance(resource, FailedReadResource):
                    built_module.insights.extend(resource.errors)
            folder.built_modules.append(built_module)

        return folder

    def _import_module(self, source: ModuleSource) -> Module:
        resources: list[ReadResource] = []
        for resource_folder, resource_files in source.resource_files_by_folder.items():
            crud_classes = RESOURCE_CRUD_BY_FOLDER_NAME.get(resource_folder)
            if not crud_classes:
                # This is handled in the module parsing phase.
                continue
            class_by_kind = {crud_class.kind: crud_class for crud_class in crud_classes}
            for resource_file in resource_files:
                resources.extend(
                    self._import_resource_file(resource_file, class_by_kind, source.variables, resource_folder)
                )

        return Module(source=source, resources=resources)

    def _import_resource_file(
        self,
        resource_file: AbsoluteFilePath,
        class_by_kind: dict[str, type[ResourceCRUD]],
        variables: list[BuildVariable],
        folder_name: str,
    ) -> list[ReadResource]:
        if "." not in resource_file.stem:
            # Todo: Discussion error or silent ignore.
            #   Reason for error is in the case were the user do not set a kind and intends to.
            #   Reason for silently ignore is that the user for example has a YAML file as part of their
            #   function code, and it is not meant to be a resource file.
            # Todo: This will fail for kinds that just endswith the kind. This is often used for
            #   for example transformation schedules.
            return []
        kind = resource_file.stem.rsplit(".", maxsplit=1)[-1]
        crud_class = class_by_kind.get(kind)
        if not crud_class:
            error = self._create_failed_read_resource_for_invalid_kind(
                resource_file, kind, folder_name, class_by_kind.keys()
            )
            return [FailedReadResource(source_path=resource_file, errors=[error])]
        error_or_string = self._read_resource_file(resource_file)
        if isinstance(error_or_string, ModelSyntaxError):
            return [FailedReadResource(source_path=resource_file, errors=[error_or_string])]
        # Content read successfully.
        if variables:
            substituted_content = self._substitute_variables_in_content(error_or_string, variables)
        else:
            substituted_content = error_or_string

        error_or_unstructured = self._parse_yaml_content(substituted_content)
        if isinstance(error_or_unstructured, ModelSyntaxError):
            return [FailedReadResource(source_path=resource_file, errors=[error_or_unstructured])]
        file_hash = calculate_hash(error_or_string, shorten=True)

        error_or_resources = self._create_resources_from_unstructured(error_or_unstructured, crud_class.yaml_cls)
        resource_type = ResourceType(resource_folder=crud_class.folder_name, kind=crud_class.kind)
        resources: list[ReadResource] = []
        for error_or_resource in error_or_resources:
            if isinstance(error_or_resource, ModelSyntaxError):
                resources.append(FailedReadResource(source_path=resource_file, errors=[error_or_resource]))
            else:
                resource, recommendations = error_or_resource
                resources.append(
                    SuccessfulReadResource(
                        source_path=resource_file,
                        resource=resource,
                        source_hash=file_hash,
                        resource_type=resource_type,
                        insights=recommendations,
                    )
                )
        return resources

    def _create_failed_read_resource_for_invalid_kind(
        self, resource_file: Path, kind: str, resource_folder: str, available_kinds: Iterable[str]
    ) -> ModelSyntaxError:
        return ModelSyntaxError(
            code="UNKNOWN-RESOURCE-KIND",
            message=f"Resource file '{resource_file.as_posix()!r}' has unknown resource kind '{kind}' for folder '{resource_folder}'",
            fix=f"Make sure the file name ends with a known resource kind for the folder. Expected kinds for folder '{resource_folder}' are: {humanize_collection(list(available_kinds))}",
        )

    def _read_resource_file(self, resource_file: Path) -> str | ModelSyntaxError:
        try:
            return safe_read(resource_file)
        except Exception as e:
            return ModelSyntaxError(
                code="RESOURCE_FILE_READ_ERROR",
                message=f"Failed to read resource file '{resource_file.as_posix()!r}': {e!s}",
                fix="Make sure the file is a valid YAML file and is accessible.",
            )

    def _substitute_variables_in_content(self, content: str, variables: list[BuildVariable]) -> str:
        raise NotImplementedError()

    def _parse_yaml_content(self, content: str) -> dict[str, Any] | list[dict[str, Any]] | ModelSyntaxError:
        try:
            return read_yaml_content(content)
        except Exception as e:
            # Todo Look for variables not replaced in the content and add fix suggestion to the error.
            #  Look for variables at an adjacent level in the YAML structure to give more specific suggestions.
            #  Jira: CDF-27203
            return ModelSyntaxError(
                code="YAML-PARSE-ERROR",
                message=f"Failed to parse YAML content: {e!s}",
                fix="Make sure the YAML content is valid.",
            )

    def _create_resources_from_unstructured(
        self, unstructured: dict[str, Any] | list[dict[str, Any]], yaml_cls: type[ToolkitResource]
    ) -> Iterable[tuple[ToolkitResource, list[Recommendation]] | ModelSyntaxError]:
        if isinstance(unstructured, dict):
            try:
                resource = yaml_cls.model_validate(unstructured, extra="forbid")
                return [(resource, [])]  # All good.
            except ValidationError as forbid_errors:
                try:
                    # Fallback to allow extra.
                    resource = yaml_cls.model_validate(unstructured, extra="ignore")
                    return [(resource, [self._create_recommendation(forbid_errors)])]
                except ValidationError:
                    return [self._create_syntax_error(forbid_errors)]
        elif isinstance(unstructured, list):
            # MyPy complains but this work.
            adapter = TypeAdapter[list[yaml_cls]](list[yaml_cls])  # type: ignore[valid-type]
            try:
                resources = adapter.validate_python(unstructured)
                return [(resource, []) for resource in resources]  # All good.
            except ValidationError as forbid_errors:
                try:
                    # Fallback to allow extra.
                    resources = adapter.validate_python(unstructured, extra="ignore")
                    recommendation = self._create_recommendation(forbid_errors)
                    # Todo: sort recommendation by resource if possible.
                    #    Jira: CDF-27204
                    return [(resource, [recommendation]) for resource in resources]
                except ValidationError:
                    return [self._create_syntax_error(forbid_errors)]
        else:
            raise NotImplementedError(
                f"Unknown unstructured YAML content. Expected a dict or list of dicts. Got {type(unstructured)}"
            )

    def _create_recommendation(self, error: ValidationError) -> Recommendation:
        # This is a quick implementation that just creates a generic recommendation for all errors. It should be extended to create more specific recommendations based on the error details.
        errors = humanize_validation_error(error)
        return Recommendation(
            code="UNKNOWN-FIELDS",
            message=f"The resource has unknown fields: {humanize_collection(errors)}",
            fix="Remove the unknown fields from the resource definition, unless you know that they are supported by the API. "
            "In that case, you can ignore this recommendation and proceed with the deployment. If you think these fields "
            "should be supported, please reach out to the Toolkit team.",
        )

    def _create_syntax_error(self, error: ValidationError) -> ModelSyntaxError:
        errors = humanize_validation_error(error)
        return ModelSyntaxError(
            code="MODEL-SYNTAX-ERROR",
            message=f"The resource definition has syntax errors: {humanize_collection(errors)}",
            fix="Make sure the resource YAML content is valid and follows the expected structure.",
        )

    def _export_module(self, module: Module, build_dir: Path) -> list[Path]:
        build_dir.mkdir(parents=True, exist_ok=True)

        built_files: list[Path] = []
        for resource in module.resources:
            if not isinstance(resource, SuccessfulReadResource):
                continue
            folder = build_dir / resource.resource_type.resource_folder
            folder.mkdir(parents=True, exist_ok=True)
            resource_file = (
                folder
                / f"resource_{sanitize_filename(str(resource.resource.as_id()))}.{resource.resource_type.kind}.yaml"
            )
            # Todo Move into Toolkit resource.
            safe_write(resource_file, yaml_safe_dump(resource.resource.model_dump(by_alias=True, exclude_unset=True)))
            built_files.append(resource_file)

        # Todo: Store source path, source hash, ID, and so on for build_linage
        return built_files

    @classmethod
    def _create_syntax_errors(cls, resource_type: ResourceType, error: ValidationError) -> Iterable[ModelSyntaxError]:
        # TODO: should be extended with humanizing of errors, this is a quick solution.

        for error_details in error.errors(include_input=True, include_url=False):
            message = error_details.get("msg", "Unknown syntax error")
            yield ModelSyntaxError(code=f"{resource_type}-SYNTAX-ERROR", message=message)

    def _local_validation(self, module: Module) -> None:
        """Local validations are post-syntax validations executed"""
        RulesOrchestrator().run(module)

    def _global_validation(self, build_folder: BuildFolder, client: ToolkitClient | None) -> None:
        """This validation is performed per resource type and not per individual resource and against CDF
        for all modules. This validation will leverage external plugins such as NEAT.
        """

        # Can be parallelized if needed
        for built_module in build_folder.built_modules:
            if not built_module.is_success:
                continue

            if files_by_resource_type := built_module.resource_by_type.get(DataModelCRUD.folder_name):
                if NeatPlugin.installed() and client and DataModelCRUD.kind in files_by_resource_type:
                    neat = NeatPlugin(client)
                    for data_model_file in files_by_resource_type[DataModelCRUD.kind]:
                        for insight in neat.validate(data_model_file.parent, data_model_file):
                            if insight not in built_module.insights:
                                built_module.insights.append(insight)

    def _write_results(self, build_folder: BuildFolder) -> None:
        return None



from typing import Literal

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject

RobotType = Literal["SPOT", "ANYMAL", "DJI_DRONE", "TAUROB", "UNKNOWN"]
MapType = Literal["WAYPOINTMAP", "THREEDMODEL", "TWODMAP", "POINTCLOUD"]


class Point3D(BaseModelObject):
    """A point in 3D space."""

    x: float
    y: float
    z: float


class Quaternion(Point3D):
    """A quaternion representing orientation."""

    w: float


class Transform(BaseModelObject):
    """Transform of the parent frame to the current frame.

    Args:
        parent_frame_external_id: The external id of the parent frame.
        translation: Transform translation (Point3D).
        orientation: Transform orientation as quaternion (Quaternion).
    """

    parent_frame_external_id: str
    translation: Point3D
    orientation: Quaternion



from collections.abc import Iterable, Sequence
from typing import Any

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.workflow_trigger import (
    WorkflowTriggerRequest,
    WorkflowTriggerResponse,
)


class WorkflowTriggersAPI(CDFResourceAPI[WorkflowTriggerResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "upsert": Endpoint(method="POST", path="/workflows/triggers", item_limit=1),
                "delete": Endpoint(method="POST", path="/workflows/triggers/delete", item_limit=1),
                "list": Endpoint(method="GET", path="/workflows/triggers", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[WorkflowTriggerResponse]:
        return PagedResponse[WorkflowTriggerResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[ExternalId]:
        return ResponseItems[ExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[WorkflowTriggerRequest]) -> list[WorkflowTriggerResponse]:
        """Create or update workflow triggers in CDF.

        Args:
            items: List of WorkflowTriggerRequest objects to create or update.
        Returns:
            List of created/updated WorkflowTriggerResponse objects.
        """
        return self._request_item_response(items, "upsert")

    # This is a duplicate of the create method, included to standardize the API interface.
    def update(self, items: Sequence[WorkflowTriggerRequest]) -> list[WorkflowTriggerResponse]:
        """Create or update workflow triggers in CDF.

        Args:
            items: List of WorkflowTriggerRequest objects to create or update.
        Returns:
            List of created/updated WorkflowTriggerResponse objects.
        """
        return self.create(items)

    def delete(self, items: Sequence[ExternalId]) -> None:
        """Delete workflow triggers from CDF.

        Args:
            items: List of ExternalId objects to delete.
        """
        self._request_no_response(items, "delete")

    def paginate(
        self,
        workflow_external_id: str | None = None,
        workflow_version: str | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[WorkflowTriggerResponse]:
        """Iterate over all workflow triggers in CDF.

        Args:
            workflow_external_id: Filter by workflow external ID.
            workflow_version: Filter by workflow version.
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of WorkflowTriggerResponse objects.
        """
        params: dict[str, Any] = {}
        if workflow_external_id:
            params["workflowExternalId"] = workflow_external_id
        if workflow_version:
            params["workflowVersion"] = workflow_version

        return self._paginate(
            cursor=cursor,
            limit=limit,
            params=params,
        )

    def iterate(
        self,
        workflow_external_id: str | None = None,
        workflow_version: str | None = None,
        limit: int = 100,
    ) -> Iterable[list[WorkflowTriggerResponse]]:
        """Iterate over all workflow triggers in CDF.

        Args:
            workflow_external_id: Filter by workflow external ID.
            workflow_version: Filter by workflow version.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of WorkflowTriggerResponse objects.
        """
        params: dict[str, Any] = {}
        if workflow_external_id:
            params["workflowExternalId"] = workflow_external_id
        if workflow_version:
            params["workflowVersion"] = workflow_version

        return self._iterate(
            limit=limit,
            params=params,
        )

    def list(
        self,
        limit: int | None = 100,
    ) -> list[WorkflowTriggerResponse]:
        """List all workflow triggers in CDF.

        Returns:
            List of WorkflowTriggerResponse objects.
        """
        return self._list(limit=limit)



from typing import Annotated, Literal

from pydantic import Field, JsonValue

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    RequestResource,
    ResponseResource,
)
from cognite_toolkit._cdf_tk.client._types import Metadata

from .identifiers import ExternalId


class TriggerRuleDefinition(BaseModelObject):
    trigger_type: str


class ScheduleTriggerRule(BaseModelObject):
    trigger_type: Literal["schedule"] = "schedule"
    cron_expression: str
    timezone: str | None = None


class DataModelingTriggerRule(BaseModelObject):
    trigger_type: Literal["dataModeling"] = "dataModeling"
    data_modeling_query: JsonValue
    batch_size: int
    batch_timeout: int


TriggerRule = Annotated[
    ScheduleTriggerRule | DataModelingTriggerRule,
    Field(discriminator="trigger_type"),
]


class NonceCredentials(BaseModelObject):
    nonce: str


class WorkflowTrigger(BaseModelObject):
    external_id: str
    trigger_rule: TriggerRule
    workflow_external_id: str
    workflow_version: str
    input: JsonValue | None = None
    metadata: Metadata | None = None

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class WorkflowTriggerRequest(WorkflowTrigger, RequestResource):
    # Note: authentication with nonce is required, but we set it to optional to
    # allow loading from file without it. This is utilized in the WorkflowTriggerCRUD.
    authentication: NonceCredentials | None = None


class WorkflowTriggerResponse(WorkflowTrigger, ResponseResource[WorkflowTriggerRequest]):
    created_time: int
    last_updated_time: int
    is_paused: bool

    @classmethod
    def request_cls(cls) -> type[WorkflowTriggerRequest]:
        return WorkflowTriggerRequest



"""Utilities for validating Python package requirements."""

import subprocess
import sys
from dataclasses import dataclass
from pathlib import Path


@dataclass
class PipValidationResult:
    """Result from validating a requirements.txt file."""

    error_message: str | None = None
    # The number of lines to include in the short error message
    max_error_lines: int = 3

    @property
    def success(self) -> bool:
        """Validation succeeded if there's no error message."""
        return self.error_message is None

    @property
    def is_credential_error(self) -> bool:
        """Check if the error appears to be related to authentication/credentials.

        Only checks for explicit HTTP authentication errors to avoid false positives
        from legitimate package not found errors.
        """
        if not self.error_message:
            return False
        credential_indicators = [
            "401",
            "403",
            "Unauthorized",
            "Forbidden",
            "Authentication",
        ]
        return any(indicator in self.error_message for indicator in credential_indicators)

    @property
    def short_error(self) -> str:
        """Get a shortened version of the error message with limited lines."""
        error_detail = self.error_message or "Unknown error"
        relevant_lines = [line for line in error_detail.strip().split("\n") if line.strip()][-self.max_error_lines :]
        error_detail = "\n      ".join(relevant_lines)
        return error_detail


def validate_requirements_with_pip(
    requirements_txt_path: Path,
    index_url: str | None = None,
    extra_index_urls: list[str] | None = None,
    timeout: int = 10,
) -> PipValidationResult:
    """Validate that requirements.txt can be resolved using pip install --dry-run.

    This simulates package installation without actually installing anything.
    It validates that:
    - All packages exist
    - All specified versions are available
    - Credentials for private repositories are valid
    - Index URLs are accessible

    Args:
        requirements_txt_path: Path to the requirements.txt file
        index_url: Optional custom package index URL (replaces PyPI)
        extra_index_urls: Optional additional package index URLs
        timeout: Timeout in seconds for the pip command

    Returns:
        PipValidationResult with success status and error details if failed

    """
    if not requirements_txt_path.exists():
        return PipValidationResult(error_message=f"Requirements file not found: {requirements_txt_path}")

    # Build pip command with custom index URLs
    args = [
        sys.executable,
        "-m",
        "pip",
        "install",
        "--dry-run",
        "--ignore-installed",
        "--no-deps",
        "--disable-pip-version-check",
        "-r",
        str(requirements_txt_path),
        *(["--index-url", index_url] if index_url else []),
        *([arg for url in (extra_index_urls or []) for arg in ["--extra-index-url", url]]),
    ]

    try:
        result = subprocess.run(args, capture_output=True, text=True, timeout=timeout, check=False)
        if result.returncode == 0:
            return PipValidationResult()
        return PipValidationResult(
            error_message=f"pip validation failed with exit code {result.returncode}:\n{result.stderr}",
        )
    except subprocess.TimeoutExpired:
        return PipValidationResult(error_message=f"pip validation timed out after {timeout} seconds")
    except (OSError, RuntimeError) as e:
        return PipValidationResult(error_message=f"Error running pip validation: {e!s}")



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
from collections.abc import Hashable, Iterable, Sequence
from typing import Any, final

from cognite.client.data_classes import capabilities
from cognite.client.data_classes.capabilities import Capability, DataSetsAcl

from cognite_toolkit._cdf_tk.client.http_client import ToolkitAPIError
from cognite_toolkit._cdf_tk.client.request_classes.filters import ClassicFilter
from cognite_toolkit._cdf_tk.client.resource_classes.dataset import DataSetRequest, DataSetResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.label import LabelRequest, LabelResponse
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitRequiredValueError,
)
from cognite_toolkit._cdf_tk.resource_classes import DataSetYAML, LabelsYAML
from cognite_toolkit._cdf_tk.utils.file import sanitize_filename

from .auth import GroupAllScopedCRUD


@final
class DataSetsCRUD(ResourceCRUD[ExternalId, DataSetRequest, DataSetResponse]):
    support_drop = False
    folder_name = "data_sets"
    resource_cls = DataSetResponse
    resource_write_cls = DataSetRequest
    yaml_cls = DataSetYAML
    kind = "DataSet"
    dependencies = frozenset({GroupAllScopedCRUD})
    _doc_url = "Data-sets/operation/createDataSets"

    @property
    def display_name(self) -> str:
        return "data sets"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[DataSetRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [DataSetsAcl.Action.Read]
            if read_only
            else [DataSetsAcl.Action.Read, DataSetsAcl.Action.Write, DataSetsAcl.Action.Owner]
        )

        return DataSetsAcl(
            actions,
            DataSetsAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: DataSetRequest | DataSetResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if not item.external_id:
            raise ToolkitRequiredValueError("DataSet must have external_id set.")
        return item.as_id()

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def as_str(cls, id: ExternalId) -> str:
        return sanitize_filename(id.external_id)

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> DataSetRequest:
        if resource.get("metadata"):
            for key, value in list(resource["metadata"].items()):
                if isinstance(value, dict | list):
                    resource["metadata"][key] = json.dumps(value)
        return DataSetRequest._load(resource)

    def dump_resource(self, resource: DataSetResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if "writeProtected" not in local and dumped.get("writeProtected") is False:
            # Default value is False, so we don't need to dump it.
            dumped.pop("writeProtected")
        if "metadata" not in local and not dumped.get("metadata"):
            # Default value is empty dict, so we don't need to dump it.
            dumped.pop("metadata", None)
        if "metadata" in dumped and "metadata" in local:
            meta_local = local["metadata"]
            for key, value in list(dumped["metadata"].items()):
                if isinstance(meta_local.get(key), dict | list):
                    try:
                        converted = json.loads(value)
                    except json.JSONDecodeError:
                        continue
                    dumped["metadata"][key] = converted

        return dumped

    def create(self, items: Sequence[DataSetRequest]) -> list[DataSetResponse]:
        return self.client.tool.datasets.create(list(items))

    def retrieve(self, ids: Sequence[ExternalId]) -> list[DataSetResponse]:
        return self.client.tool.datasets.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[DataSetRequest]) -> list[DataSetResponse]:
        return self.client.tool.datasets.update(list(items), mode="replace")

    def delete(self, ids: Sequence[ExternalId]) -> int:
        raise NotImplementedError("CDF does not support deleting data sets.")

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[DataSetResponse]:
        for items in self.client.tool.datasets.iterate(limit=None):
            yield from items


@final
class LabelCRUD(ResourceCRUD[ExternalId, LabelRequest, LabelResponse]):
    folder_name = "classic"
    resource_cls = LabelResponse
    resource_write_cls = LabelRequest
    yaml_cls = LabelsYAML
    kind = "Label"
    dependencies = frozenset({DataSetsCRUD, GroupAllScopedCRUD})
    _doc_url = "Labels/operation/createLabelDefinitions"
    support_update = False

    @property
    def display_name(self) -> str:
        return "labels"

    @classmethod
    def get_id(cls, item: LabelRequest | LabelResponse | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if not item.external_id:
            raise ToolkitRequiredValueError("Label must have external_id set.")
        return item.as_id()

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[LabelRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        scope: capabilities.LabelsAcl.Scope.All | capabilities.LabelsAcl.Scope.DataSet = (  # type: ignore[valid-type]
            capabilities.LabelsAcl.Scope.All()
        )
        if items:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = capabilities.LabelsAcl.Scope.DataSet(list(data_set_ids))

        actions = (
            [capabilities.LabelsAcl.Action.Read]
            if read_only
            else [capabilities.LabelsAcl.Action.Read, capabilities.LabelsAcl.Action.Write]
        )

        return capabilities.LabelsAcl(actions, scope)

    def create(self, items: Sequence[LabelRequest]) -> list[LabelResponse]:
        return self.client.tool.labels.create(list(items))

    def retrieve(self, ids: Sequence[ExternalId]) -> list[LabelResponse]:
        return self.client.tool.labels.retrieve(list(ids), ignore_unknown_ids=True)

    def delete(self, ids: Sequence[ExternalId]) -> int:
        if not ids:
            return 0
        try:
            self.client.tool.labels.delete(list(ids))
        except ToolkitAPIError as e:
            if missing := {ExternalId.model_validate(item) for item in e.missing or []}:
                if existing := (set(ids) - missing):
                    self.client.tool.labels.delete(list(existing))
                    return len(existing)
                else:
                    return 0
            raise
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[LabelResponse]:
        filter: ClassicFilter | None = None
        if data_set_external_id is not None:
            filter = ClassicFilter(data_set_ids=[ExternalId(external_id=data_set_external_id)])
        for items in self.client.tool.labels.iterate(filter=filter):
            yield from items

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, ExternalId(external_id=item["dataSetExternalId"])

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> LabelRequest:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        return LabelRequest._load(resource)

    def dump_resource(self, resource: LabelResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        return dumped



import builtins
from typing import Any, ClassVar, Literal

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)
from cognite_toolkit._cdf_tk.client._types import Metadata

from .identifiers import ExternalId, InternalOrExternalId
from .instance_api import NodeReference


class TimeSeries(BaseModelObject):
    external_id: str | None = None
    name: str | None = None
    is_string: bool = False
    metadata: Metadata | None = None
    unit: str | None = None
    unit_external_id: str | None = None
    asset_id: int | None = None
    is_step: bool = False
    description: str | None = None
    security_categories: list[int] | None = None
    data_set_id: int | None = None

    def as_id(self) -> InternalOrExternalId:
        if self.external_id is None:
            raise ValueError("Cannot convert TimeSeriesRequest to ExternalId when external_id is None")
        return ExternalId(external_id=self.external_id)


class TimeSeriesRequest(TimeSeries, UpdatableRequestResource):
    container_fields: ClassVar[frozenset[str]] = frozenset({"metadata", "security_categories"})
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"is_step"})

    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, Any]:
        dumped = super().as_update(mode)
        update = dumped["update"] if "update" in dumped else dumped
        # isString is immutable in CDF, so we remove it from update payloads
        update.pop("isString", None)
        return dumped


class TimeSeriesResponse(TimeSeries, ResponseResource[TimeSeriesRequest]):
    id: int
    instance_id: NodeReference | None = None
    pending_instance_id: NodeReference | None = None
    type: str
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> builtins.type[TimeSeriesRequest]:
        return TimeSeriesRequest
